import librosa
import torch
import torchaudio.functional as F
import torchaudio.transforms as T
import numpy as np
from scipy.signal import lfilter
import dataclasses as dtc

__all__ = [
    'FileToSignal',
    'Normalize',
    'Emphasis',
    'Deemphasis',
    'MuLawCompress',
    'MuLawExpand',
    'STFT',
    'ISTFT',
    'MagSpec',
    'GLA'
]

N_FFT = 2048
HOP_LENGTH = 512
SR = 22050
Q_LEVELS = 256


class FModule:
    """
    base class for implementing message passing in callable objects
    """

    @property
    def functions(self) -> dict:
        raise NotImplementedError

    def __call__(self, inputs):
        return self.functions[type(inputs)](inputs)


@dtc.dataclass
class FileToSignal(FModule):
    sr: int = SR

    @property
    def functions(self):
        return {
            str: lambda path: librosa.load(path, sr=self.sr)[0]
        }


@dtc.dataclass
class Normalize(FModule):
    p: int = 1
    dim: int = -1

    @property
    def functions(self):
        def np_func(inputs):
            return librosa.util.normalize(inputs, norm=self.p, axis=self.dim)

        def torch_func(inputs):
            return torch.nn.functional.normalize(inputs, p=self.p, dim=self.dim)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class Emphasis(FModule):
    emphasis: float = 0.

    @property
    def functions(self):
        def np_func(inputs):
            return lfilter([1, -self.emphasis], [1], inputs)

        def torch_func(inputs):
            return F.lfilter(inputs,
                             torch.tensor([1, 0]).to(inputs),  # a0, a1
                             torch.tensor([1, -self.emphasis]).to(inputs))  # b0, b1

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class Deemphasis(FModule):
    emphasis: float = 0.

    @property
    def functions(self):
        def np_func(inputs):
            return lfilter([1 - self.emphasis], [1, -self.emphasis], inputs)

        def torch_func(inputs):
            return F.lfilter(inputs,
                             torch.tensor([1, - self.emphasis]).to(inputs),  # a0, a1
                             torch.tensor([1 - self.emphasis, 0]).to(inputs))  # b0, b1

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class MuLawCompress(FModule):
    q_levels: int = Q_LEVELS

    @property
    def functions(self):
        def np_func(inputs):
            qx = librosa.mu_compress(inputs, self.q_levels - 1, quantize=True)
            qx = qx + self.q_levels // 2
            return qx

        def torch_func(inputs):
            mod = T.MuLawEncoding(self.q_levels)
            return mod(inputs)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class MuLawExpand(FModule):
    q_levels: int = Q_LEVELS

    @property
    def functions(self):
        def np_func(inputs):
            return librosa.mu_expand(inputs - self.q_levels // 2, self.q_levels - 1, quantize=True)

        def torch_func(inputs):
            mod = T.MuLawDecoding(self.q_levels)
            return mod(inputs)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class STFT(FModule):
    n_fft: int = N_FFT
    hop_length: int = HOP_LENGTH

    @property
    def functions(self):
        def np_func(inputs):
            # returned shape is (time x freq)
            return librosa.stft(inputs, n_fft=self.n_fft, hop_length=self.hop_length).T

        def torch_func(inputs):
            mod = T.Spectrogram(self.n_fft, hop_length=self.hop_length, power=1.,
                                wkwargs=dict(device=inputs.device))
            # returned shape is (..., time x freq)
            return mod(inputs).transpose(-1, -2).contiguous()

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class ISTFT(FModule):
    n_fft: int = N_FFT
    hop_length: int = HOP_LENGTH

    @property
    def functions(self):
        def np_func(inputs):
            # inputs is of shape (time x freq)
            return librosa.istft(inputs.T, n_fft=self.n_fft, hop_length=self.hop_length, )

        def torch_func(inputs):
            # inputs is of shape (time x freq)
            y = torch.istft(inputs.transpose(-1, -2).contiguous(),
                            n_fft=self.n_fft, hop_length=self.hop_length,
                            window=torch.hann_window(self.n_fft, device=inputs.device))
            return y

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class MagSpec(STFT):

    @property
    def functions(self):
        sup_f = super(MagSpec, self).functions
        # dict comprehension would result in a single function for
        # all types, so we declare the dict manually...
        return {
            np.ndarray: lambda x: abs(sup_f[np.ndarray](x)),
            torch.Tensor: lambda x: abs(sup_f[torch.Tensor](x))
        }


@dtc.dataclass
class GLA(FModule):
    n_fft: int = N_FFT
    hop_length: int = HOP_LENGTH
    n_iter: int = 32

    @property
    def functions(self):
        def np_func(inputs):
            # inputs is of shape (time x freq)
            return librosa.griffinlim(inputs.T, hop_length=self.hop_length, n_iter=self.n_iter)

        def torch_func(inputs):
            gla = T.GriffinLim(n_fft=self.n_fft, hop_length=self.hop_length, power=1.,
                               wkwargs=dict(device=inputs.device))
            # inputs is of shape (time x freq)
            return gla(inputs.transpose(-1, -2).contiguous())

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }



import dataclasses as dtc
from inspect import signature

import pytorch_lightning as pl


def _filter_cls_kwargs(cls, kwargs):
    valids = signature(cls.__init__).parameters.keys()
    return {k: v for k, v in kwargs.items() if k in valids}


def model(cls):
    """

    Parameters
    ----------
    cls

    Returns
    -------

    """

    bases = cls.__bases__
    bases = [cls, *bases]

    # get the __init__ signatures of the bases
    sigs = [dict(signature(d.__init__).parameters) for d in bases
            if d is not pl.LightningModule]
    # collect positional arguments for validation
    posits = {k: v for s in sigs for k, v in s.items() if v.default is v.empty}
    if any(k for k in posits.keys() if k != "self"):
        for b in bases:
            posits = {k: v for k, v in dict(signature(b.__init__).parameters).items() if v.default is v.empty}
            if any(k for k in posits.keys() if k != "self"):
                raise ValueError("All dependencies should have kwargs only."
                                 " Got following positional args: " + str([k for k in posits.keys() if k != "self"]) +\
                                 "for the following parent class: " + str(b))

    # collect keywords arguments
    kws = {k: v for s in sigs for k, v in s.items() if v.default is not v.empty}

    # build fields for make_dataclass
    fields = [(k, type(v.default), dtc.field(default=v.default)) for k, v in kws.items()]

    # method to initialize all the bases
    def __post_init__(self):
        super(pl.LightningModule, self).__init__()
        for b in bases:
            if hasattr(b, '__post_init__'):
                # avoid recursion if b is a dataclass already
                b.__post_init__(self)
            else:
                b.__init__(self, **_filter_cls_kwargs(b, self.__dict__))

        # set hparams manually since lightning will not be able to infer them...
        self._set_hparams({k: getattr(self, k) for k in self.__args__})

    # namespace of the class
    ns = dict(cls.__dict__)
    ns.update({
        "__post_init__": __post_init__,
        '__args__': tuple(kws.keys()),
    })
    return dtc.make_dataclass(cls.__qualname__,
                              fields,
                              bases=(*(b for b in bases if b is not cls),
                                     *((pl.LightningModule,) if pl.LightningModule not in bases else tuple())),
                              namespace=ns,
                              init=True, repr=False, eq=False, frozen=False, unsafe_hash=True
                              )


class Model(type):
    def __new__(mcs, name, bases, namespace):
        cls = super(Model, mcs).__new__(mcs, name, bases, namespace)
        return model(cls)


import dataclasses as dtc

__all__ = [
    'Feature'
]


@dtc.dataclass
class Feature:
    """
    class for centralizing methods pertaining to a particular type, class or format of data.
    """
    __ext__ = None

    @property
    def dim(self):
        """
        Returns
        -------
        dim: int, tuple
            information about the shape / dimensionality of the feature
        """
        raise NotImplementedError

    @property
    def params(self):
        """
        Returns
        -------
        params: dict
            the parameters of this Feature
        """
        return dtc.asdict(self)

    @property
    def encoders(self):
        """
        Returns
        -------
        encoders: dict
            dict or mapping where keys are types and values callables taking only one argument
        """
        class AnyIdentity:
            def __getitem__(self, item):
                return lambda x: x
        return AnyIdentity()

    @property
    def decoders(self):
        """
        Returns
        -------
        decoders: dict
            dict or mapping where keys are types and values callables taking only one argument
        """
        class AnyIdentity:
            def __getitem__(self, item):
                return lambda x: x

        return AnyIdentity()

    def __call__(self, inputs, encode=True):
        """
        only apply the layer of encoder/decoder owned by this class. Use it for augmentation, transformation.

        Parameters
        ----------
        inputs

        encode: bool, opt
            apply self.encoders if True, self.decoders otherwise

        Returns
        -------

        """
        funcs = self.encoders if encode else self.decoders
        return funcs[type(inputs)](inputs)

    def encode(self, inputs):
        """
        apply all the layers of encoders inherited by this class to `inputs`.

        Parameters
        ----------
        inputs

        Returns
        -------

        """
        if hasattr(super(), 'encoders'):
            inputs = super().encoders[type(inputs)](inputs)
        return self.encoders[type(inputs)](inputs)

    def decode(self, inputs):
        """
        apply all the layers of decoders inherited by this class to `inputs`.

        Parameters
        ----------
        inputs

        Returns
        -------

        """
        inputs = self.decoders[type(inputs)](inputs)
        if hasattr(super(), 'decoders'):
            inputs = super().decoders[type(inputs)](inputs)
        return inputs

    def load(self, path):
        """
        reads the file at `path` and returns it as numpy array

        Parameters
        ----------
        path

        Returns
        -------

        """
        # since this is optional, we don't raise any exception
        pass

    def post_create(self, db, schema_key):
        # since this is optional, we don't raise any exception
        pass

    def play(self, inputs):
        raise NotImplementedError

    def display(self, inputs):
        raise NotImplementedError

    def write(self, filename, inputs):
        raise NotImplementedError



from pytorch_lightning import LightningModule
from typing import Iterable
import dataclasses as dtc
from torch.utils.data import Dataset


from ...data import Database, DataModule

__all__ = [
    'IData'
]


class IData(LightningModule):

    def __init__(self):
        LightningModule.__init__(self)

    @property
    def dm(self) -> DataModule:
        return self.trainer.data_connector.datamodule

    @classmethod
    def schema(cls, *args, **kwargs) -> dict:
        raise NotImplementedError

    @classmethod
    def dependant_hp(cls, db: Database) -> dict:
        return {}

    def batch_signature(self, stage):
        raise NotImplementedError

    def loader_kwargs(self, stage, datamodule) -> dict:
        raise NotImplementedError



from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from pytorch_lightning.trainer.states import TrainerState
from typing import Iterable
import os

__all__ = [
    'MMKCheckpoint'
]


class MMKCheckpoint(ModelCheckpoint):
    def __init__(self,
                 dirpath="./",
                 epochs=None,
                 ):
        filename = "{step}" if type(epochs) is float else "{epoch}"
        self.epochs = epochs

        super(MMKCheckpoint, self).__init__(
            monitor=None,
            verbose=False,
            save_last=None,
            save_top_k=None,
            save_weights_only=False,  # we just save optimizers separately
            mode="min",
            period=1,
            dirpath=os.path.join(dirpath, "states"),
            filename=filename
        )

    def should_save(self, epoch, step):
        if type(self.epochs) is int:
            return epoch > 0 and (epoch % self.epochs) == 0
        elif type(self.epochs) is float:  # Doesn't work for now...
            return (step % int(self.epochs)) == 0
        elif isinstance(self.epochs, Iterable):
            return epoch in self.epochs
        return False

    def save_checkpoint(self, trainer, pl_module):
        epoch, global_step = trainer.current_epoch + 1, trainer.global_step
        if trainer.state == TrainerState.INTERRUPTED or \
                epoch == trainer.max_epochs or self.should_save(epoch, global_step):
            # hacks the code from github
            self._add_backward_monitor_support(trainer)
            self._validate_monitor_key(trainer)

            # track epoch when ckpt was last checked
            self.last_global_step_saved = global_step

            filepath = self.format_checkpoint_name(epoch, global_step, {})

            self._save_model(filepath, trainer, pl_module)

    def on_save_checkpoint(self, trainer, model, checkpoint):
        """returns the state to be saved in the checkpoint"""
        return {"epochs": self.epochs}

    def on_load_checkpoint(self, state):
        """update own state according to checkpoint"""
        self.epochs = state["epochs"]



from pytorch_lightning.callbacks.progress import Callback
from IPython import get_ipython

__all__ = [
    'is_notebook',
    'EpochProgressBarCallback',
    'GradNormCallback',
    'tqdm'
]


def is_notebook():
    shell = get_ipython().__class__.__name__
    if shell in ('ZMQInteractiveShell', "Shell"):
        # local and colab notebooks
        return True
    elif shell == 'TerminalInteractiveShell':
        return False
    else:
        return False


if is_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


class EpochProgressBarCallback(Callback):
    def on_train_start(self, trainer, pl_module):
        self.epoch_bar = tqdm(range(1, trainer.max_epochs), unit="epoch",
                              position=0, leave=False, dynamic_ncols=True)

    def on_train_epoch_end(self, trainer, *args):
        if trainer.val_dataloaders is not None and any(trainer.val_dataloaders):
            return
        else:
            self.epoch_bar.update()

    def on_validation_epoch_end(self, *args):
        self.epoch_bar.update()


class GradNormCallback(Callback):

    def __init__(self):
        self.gradnorms = []

    def on_after_backward(self, trainer, pl_module: 'LightningModule') -> None:
        self.gradnorms += [pl_module.grad_norm(1.)]


import torch
import numpy as np
import librosa
import dataclasses as dtc
import IPython.display as ipd
import soundfile as sf

from ..data import Feature
from . import fmodules as T

__all__ = [
    'AudioSignal',
    'MuLawSignal',
    'Spectrogram'
]


@dtc.dataclass
class AudioSignal(Feature):
    """
    audio signal managers
    """
    __ext__ = 'audio'

    sr: int = 22050
    normalize: bool = True
    emphasis: float = 0.

    @property
    def dim(self):
        return 1

    @property
    def encoders(self):
        def func(inputs):
            if self.emphasis:
                inputs = T.Emphasis(self.emphasis)(inputs)
            if self.normalize:
                inputs = T.Normalize()(inputs)
            return inputs

        return {
            np.ndarray: func,
            torch.Tensor: func
        }

    @property
    def decoders(self):
        def func(inputs):
            if self.emphasis:
                inputs = T.Deemphasis(self.emphasis)(inputs)
            if self.normalize:
                inputs = T.Normalize()(inputs)
            return inputs

        return {
            np.ndarray: func,
            torch.Tensor: func
        }

    def load(self, path):
        y = T.FileToSignal(self.sr)(path)
        y = self.encode(y)
        return y

    def display(self, inputs, **waveplot_kwargs):
        waveplot_kwargs.setdefault('sr', self.sr)
        return librosa.display.waveplot(inputs, **waveplot_kwargs)

    def play(self, inputs):
        return ipd.display(ipd.Audio(inputs, rate=self.sr))

    def write(self, filename, inputs):
        return sf.write(filename, inputs, self.sr, 'PCM_24')


@dtc.dataclass(unsafe_hash=True)
class MuLawSignal(AudioSignal):
    q_levels: int = 256

    @property
    def dim(self):
        return self.q_levels

    @property
    def encoders(self):
        return {
            np.ndarray: T.MuLawCompress(q_levels=self.q_levels),
            torch.Tensor: T.MuLawCompress(q_levels=self.q_levels)
        }

    @property
    def decoders(self):
        return {
            np.ndarray: T.MuLawExpand(self.q_levels),
            torch.Tensor: T.MuLawExpand(self.q_levels)
        }


@dtc.dataclass(unsafe_hash=True)
class Spectrogram(AudioSignal):
    n_fft: int = 2048
    hop_length: int = 512
    magspec: bool = False

    @property
    def dim(self):
        return self.n_fft // 2 + 1

    @property
    def encoders(self):
        if self.magspec:
            return {
                np.ndarray: T.MagSpec(self.n_fft, self.hop_length),
                torch.Tensor: T.MagSpec(self.n_fft, self.hop_length)
            }
        return {
            np.ndarray: T.STFT(self.n_fft, self.hop_length),
            torch.Tensor: T.STFT(self.n_fft, self.hop_length)
        }

    @property
    def decoders(self):
        if self.magspec:
            return {
                np.ndarray: T.GLA(self.n_fft, self.hop_length, n_iter=32),
                torch.Tensor: T.GLA(self.n_fft, self.hop_length, n_iter=32)
            }
        return {
            np.ndarray: T.ISTFT(self.n_fft, self.hop_length),
            torch.Tensor: T.ISTFT(self.n_fft, self.hop_length)
        }

    def display(self, inputs, **waveplot_kwargs):
        y = self.decode(inputs)
        return super(Spectrogram, self).display(y, **waveplot_kwargs)

    def play(self, inputs):
        y = self.decode(inputs)
        return super(Spectrogram, self).play(y)

    def write(self, filename, inputs):
        y = self.decode(inputs)
        return super(Spectrogram, self).write(filename, y)



import math

import torch
from torch.utils.data import Sampler, RandomSampler, BatchSampler


__all__ = [
    'TBPTTSampler'
]


class TBPTTSampler(Sampler):
    """
    yields batches of indices for performing Truncated Back Propagation Through Time
    """

    def __init__(self,
                 n_samples,
                 batch_size=64,  # nbr of "tracks" per batch
                 chunk_length=8 * 16000,  # total length of a track
                 seq_len=512  # nbr of samples per backward pass
                 ):
        super().__init__(None)
        self.n_samples = n_samples
        self.chunk_length = chunk_length
        self.seq_len = seq_len
        self.n_chunks = self.n_samples // self.chunk_length
        self.remainder = self.n_samples % self.chunk_length
        self.n_per_chunk = self.chunk_length // self.seq_len
        self.batch_size = min(batch_size, self.n_chunks)

    def __iter__(self):
        smp = RandomSampler(torch.arange(self.n_chunks))
        for top in BatchSampler(smp, self.batch_size, False):  # don't drop last!
            offsets = torch.randint(0, self.remainder, (self.batch_size,))
            top = tuple(o + (t * self.chunk_length) for t, o in zip(top, offsets))
            for start in range(self.n_per_chunk):
                # start indices of the batch
                yield tuple(t + (start * self.seq_len) for t in top)

    def __len__(self):
        return int(max(1, math.floor(self.n_chunks / self.batch_size)) * self.n_per_chunk)


from .callbacks import *
from .checkpoint import *
from .idata import *
from .hooks import *
from .loggers import *
from .loss_functions import *
from .optim import *
from .sequence_model import *


__all__ = [_ for _ in dir() if not _.startswith("_")]


import torch
import torch.nn as nn
from typing import Optional
from dataclasses import dataclass

from ..networks.parametrized_gaussian import ParametrizedGaussian

__all__ = [
    'EncoderLSTM',
    'DecoderLSTM',
    'Seq2SeqLSTM',
]


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class EncoderLSTM(nn.Module):
    input_d: int
    model_dim: int
    num_layers: int
    n_lstm: Optional[int] = 1
    bottleneck: Optional[str] = "add"
    n_fc: Optional[int] = 1
    bias: Optional[bool] = False
    weight_norm: Optional[bool] = False

    def __postinit__(self):
        nn.Module.__init__(self)
        self.lstms = nn.ModuleList([
            nn.LSTM(self.input_d if i == 0 else self.model_dim,
                    self.model_dim if self.bottleneck == "add" else self.model_dim // 2,
                    bias=self.bias,
                    num_layers=self.num_layers,
                    batch_first=True, bidirectional=True)
            for i in range(self.n_lstm)
        ])
        self.fc = nn.Sequential(
            *[nn.Sequential(nn.Linear(self.model_dim, self.model_dim), nn.Tanh()) for _ in range(self.n_fc - 1)],
            nn.Linear(self.model_dim, self.model_dim, bias=False),  # NO ACTIVATION !
        )
        if self.weight_norm:
            for name, p in dict(self.lstms.named_parameters()).items():
                if "weight" in name:
                    torch.nn.utils.weight_norm(self.lstms, name)

    def forward(self, x, h0=None, c0=None):
        if h0 is None or c0 is None:
            hidden = tuple()
        else:
            hidden = (h0, c0)
        ht, ct = None, None
        for i, lstm in enumerate(self.lstms):
            out, (ht, ct) = lstm(x, *hidden)
            # sum forward and backward nets
            out = out.view(*out.size()[:-1], self.model_dim, 2).sum(dim=-1)
            # take residuals AFTER the first lstm
            x = out if i == 0 else x + out
        states = self.first_and_last_states(x)
        return self.fc(states), (ht, ct)

    def first_and_last_states(self, sequence):
        first_states = sequence[:, 0, :]
        last_states = sequence[:, -1, :]
        if self.bottleneck == "add":
            return first_states + last_states
        else:
            return torch.cat((first_states, last_states), dim=-1)


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class DecoderLSTM(nn.Module):
    model_dim: int
    num_layers: int
    bottleneck: Optional[str] = "add"
    bias: Optional[bool] = False
    weight_norm: Optional[tuple] = (False, False)

    def __postinit__(self):
        nn.Module.__init__(self)
        self.lstm1 = nn.LSTM(self.model_dim, self.model_dim if self.bottleneck == "add" else self.model_dim // 2,
                             bias=self.bias,
                             num_layers=self.num_layers, batch_first=True, bidirectional=True)
        self.lstm2 = nn.LSTM(self.model_dim, self.model_dim if self.bottleneck == "add" else self.model_dim // 2,
                             bias=self.bias,
                             num_layers=self.num_layers, batch_first=True, bidirectional=True)
        for lstm, wn in zip([self.lstm1, self.lstm2], self.weight_norm):
            if wn:
                for name, p in dict(lstm.named_parameters()).items():
                    if "weight" in name:
                        torch.nn.utils.weight_norm(lstm, name)

    def forward(self, x, hiddens, cells):
        if hiddens is None or cells is None:
            output, (_, _) = self.lstm1(x)
        else:
            # ALL decoders get hidden states from encoder
            output, (_, _) = self.lstm1(x, (hiddens, cells))
        # sum forward and backward nets
        output = output.view(*output.size()[:-1], self.model_dim, 2).sum(dim=-1)

        if hiddens is None or cells is None:
            output2, (hiddens, cells) = self.lstm2(output)
        else:
            output2, (hiddens, cells) = self.lstm2(output, (hiddens, cells))
        # sum forward and backward nets
        output2 = output2.view(*output2.size()[:-1], self.model_dim, 2).sum(dim=-1)

        # sum the outputs
        return output + output2, (hiddens, cells)


class Seq2SeqLSTM(nn.Module):
    def __init__(self,
                 input_dim,
                 model_dim,
                 num_layers=1,
                 n_lstm=1,
                 bottleneck="add",
                 n_fc=1):
        super(Seq2SeqLSTM, self).__init__()
        self.enc = EncoderLSTM(input_dim, model_dim, num_layers, n_lstm, bottleneck, n_fc)
        self.dec = DecoderLSTM(model_dim, num_layers, bottleneck)
        self.sampler = ParametrizedGaussian(model_dim, model_dim)
        self.fc_out = nn.Linear(model_dim, input_dim, bias=False)

    def forward(self, x, output_length=None):
        coded, (h_enc, c_enc) = self.enc(x)
        if output_length is None:
            output_length = x.size(1)
        coded = coded.unsqueeze(1).repeat(1, output_length, 1)
        residuals, _, _ = self.sampler(coded)
        coded = coded + residuals
        output, (_, _) = self.dec(coded, h_enc, c_enc)
        return self.fc_out(output).abs()



from dataclasses import dataclass
from functools import wraps

import torch.nn as nn
import torch
from copy import deepcopy

__all__ = [
    'HOM',
    'HOMSequential',
    'SequentialAdd',
    'SequentialMul',
    'SequentialCollect',
    'Paths',
    'FPaths',
    'AddPaths',
    'MulPaths',
    'block',
    'Tiers',
    'Skips',
    'GatedUnit'
]


# *******************************************************************
# ************************ HOM **************************************
# *******************************************************************


class NotAModule(nn.Module):
    """
    the nn.Module equivalent of None
    """
    pass


class HOM(nn.Sequential, nn.ModuleList):

    @staticmethod
    def replace_none(module):
        return module if module is not None else NotAModule()

    @staticmethod
    def output_or_none(mod, x):
        if type(mod) is NotAModule or x is None:
            return None
        return mod(x)

    @staticmethod
    def flat_inputs(*inputs):
        return inputs[0] if len(inputs) == 1 and isinstance(inputs[0], tuple) else inputs

    def __init__(self, *modules):
        super().__init__(*(self.replace_none(m) for m in modules))

    def forward(self, *inputs):
        raise NotImplementedError


# *******************************************************************
# ******************* HOMSequential *********************************
# *******************************************************************


class HOMSequential(HOM):
    ops_ = None

    def __init__(self, *modules):
        nn.Sequential.__init__(self, *[self.replace_none(m) for m in modules])

    def forward(self, input):
        out = None
        for mod in self:
            out = self.output_or_none(mod, input)
            if out is not None:
                if self.ops_ is not None:
                    self.ops_(input, out)
                else:
                    input = out
        return out


class SequentialAdd(HOMSequential):
    @property
    def ops_(self):
        return torch.Tensor.add_


class SequentialMul(HOMSequential):
    @property
    def ops_(self):
        return torch.Tensor.mul_


class SequentialCollect(HOMSequential):

    def forward(self, input):
        collected = []
        for mod in self:
            out = self.output_or_none(mod, input)
            if out is not None:
                if self.ops_ is not None:
                    self.ops_(input, out)
                else:
                    input = out
            collected += [out]
        return tuple(collected)


# *******************************************************************
# ************************* Paths ***********************************
# *******************************************************************


class Paths(HOM):

    def __init__(self, *modules):
        super(Paths, self).__init__(*modules)

    def forward(self, *inputs):
        inputs = self.flat_inputs(*inputs)
        return tuple(self.output_or_none(mod, x) for mod, x in zip(self, inputs))


class FPaths(Paths):
    @property
    def ops(self):
        return None

    def forward(self, *inputs):
        # flatten ((x, x, ....), ) to (x, x, ...)
        inputs = self.flat_inputs(*inputs)
        inputs = (x for x in inputs)
        modules = (m for m in self)
        out = None
        # if the firsts modules were Nones
        while out is None:
            out = self.output_or_none(next(modules), (next(inputs)))
        # now accum the outputs
        for mod, x in zip(modules, inputs):
            output = self.output_or_none(mod, x)
            if output is not None:
                # all output shapes have to be the same!...
                self.ops(out, output)
        return out


class AddPaths(FPaths):
    @property
    def ops(self):
        return torch.Tensor.add


class MulPaths(FPaths):
    @property
    def ops(self):
        return torch.Tensor.mul


# *******************************************************************
# ************************ block ************************************
# *******************************************************************


def block(name, layer_func, container_cls=HOMSequential):
    def init(self, n_layers, *args, **kwargs):
        container_cls.__init__(self, *(layer_func(i, *args, **kwargs)
                                       for sub_block in n_layers for i in range(sub_block)))

    return type(name, (container_cls,), {"__init__": init})


class Tiers(HOMSequential):

    def forward(self, tiers_inputs):
        output = None
        for tier, inpt in zip(self, tiers_inputs):
            output = tier(inpt, output)
        return output


# *******************************************************************
# ************************* MISC ************************************
# *******************************************************************

class Skips(nn.Module):
    def __init__(self, skips):
        super().__init__()
        self.skp = skips

    def forward(self, x, skips=None):
        h = self.skp(x)
        if skips is None:
            skips = torch.zeros_like(h, device=h.device)
        skips.add_(h)
        return skips


class GatedUnit(MulPaths):

    def __new__(cls, module):
        class GU(MulPaths):
            def forward(self, inputs):
                return super(MulPaths, self).forward(inputs, inputs)

        return GU(nn.Sequential(module, nn.Tanh()),
                  nn.Sequential(deepcopy(module), nn.Sigmoid()))


def propped(*bases):
    def wrapper(cls):
        cls = dataclass(cls, init=True, repr=False, eq=False, frozen=False)

        @wraps(cls.__init__)
        def new_init(self, *args, **kwargs):
            for base in bases:
                base.__init__(self)
            cls.__init__(self, *args, **kwargs)

        cls.__init__ = new_init
        return cls

    return wrapper



from functools import partial
import h5py
import torch
from torch.utils.data._utils.collate import default_convert
from torch.utils.data import random_split, Dataset
import numpy as np
import pandas as pd
import os
from argparse import Namespace
import pytorch_lightning as pl

from .create import make_root_db, write_feature
from .feature import Feature
from .regions import Regions
from ..file_walker import FileWalker, EXTENSIONS

__all__ = [
    "Database"
]


class FeatureProxy(object):
    """
    Interface to the numerical data (array) stored in a .h5 file created with ``mimikit.data.make_root_db``

    ``FeatureProxy`` objects are like ``numpy`` arrays (on disk) and can be used as map-style ``Dataset``
    because they implement ``__len__`` and ``__getitem__``.

    Parameters
    ----------
    h5_file : str
        name of the file
    ds_name : str
        name of the ``h5py.Dataset`` containing the data to be served

    Attributes
    ----------
    N : int
        the length of the first dimension of the underlying array
    shape : tuple of int
        the shape of the underlying array
    attrs : dict
        dictionary of additional information about the data as returned
        by the ``extract_func`` passed to ``make_root_db``.
        Typically, this is where you want to store the parameters of your extracting function, e.g.
        sample rate, hop length etc.
    files : Regions
        pandas DataFrame where each row contains name, start, stop & duration of one stored file.
        see ``Regions`` for more information
    regions : Regions or None
        pandas DataFrame containing index, start, stop & duration for segments ('sub-regions' of the files)
        This attribute isn't None only if the ``extract()`` method of the parent ``Database`` class returned such
        an object for this feature.
    """

    def __init__(self, h5_file: str, ds_name: str, keep_open=False):

        self.h5_file = h5_file
        self.name = ds_name
        with h5py.File(h5_file, "r") as f:
            ds = f[self.name]
            self.N = ds.shape[0]
            self.shape = ds.shape
            self.dtype = ds.dtype
            self.attrs = {k: v for k, v in ds.attrs.items()}
            has_files = self.name + "_files" in f
            has_regions = self.name + "_regions" in f
        self.files = Regions(pd.read_hdf(h5_file, self.name + "_files", mode="r")) if has_files else None
        self.regions = Regions(pd.read_hdf(h5_file, self.name + "_regions", mode="r")) if has_regions else None
        # handle to the file when keeping open. To support torch's Dataloader, we have to open the file by the
        # first getitem request
        self._f = None
        self.keep_open = keep_open

    def __len__(self):
        return self.N

    def __getitem__(self, item):
        if self.keep_open:
            if self._f is None:
                self._f = h5py.File(self.h5_file, "r+")
            return self._f[self.name][item]
        with h5py.File(self.h5_file, "r") as f:
            rv = f[self.name][item]
        return rv

    def __setitem__(self, item, value):
        if self.keep_open:
            if self._f is None:
                self._f = h5py.File(self.h5_file, "r+")
            self._f[self.name][item] = value
        with h5py.File(self.h5_file, "r+") as f:
            f[self.name][item] = value
        return

    def get_regions(self, regions):
        """
        get the data (numpy array) corresponding to the rows of `regions`

        Parameters
        ----------
        regions : Regions
            the files you want to get as array

        Returns
        -------
        data : np.ndarray
            the concatenated data for all the requested files

        Examples
        --------

            .. code-block::

                from mimikit import Database
                X = Database("my-db.h5").fft
                # only get files 0, 3 & 7
                X.get_regions(X.regions.iloc[[0, 3, 7]])
        """
        slices = regions.slices(0)
        return np.concatenate(tuple(self[slice_i] for slice_i in slices), axis=0)

    def close(self):
        if self._f is not None:
            self._f.close()
        return self

    def __repr__(self):
        return "<FeatureProxy: '%s/%s'>" % (self.h5_file, self.name)


class Database(Dataset):
    """
    interface to .h5 databases created by mimikit

    Parameters
    ----------
    h5_file : str
        path to the .h5 file containing the data
    keep_open : bool, optional
        whether to keep the h5 file open or close it after each query.
        Default is ``False``.

    Attributes
    ----------
    <feature_proxy> : FeatureProxy
        each feature created by the extracting function passed to ``make_root_db``
        is automatically added as attribute. If the extracting function returned a feature
        by the name ``"fft"``, the attribute ``fft`` of type ``FeatureProxy`` will be automatically
        added when the file is loaded and you will be able to access it through ``db.fft``.
    """

    schema = {}

    def __init__(self, h5_file: str, keep_open=False):
        self.h5_file = h5_file
        self.keep_open = keep_open
        with h5py.File(h5_file, "r") as f:
            self.attrs = {k: v for k, v in f.attrs.items()}
            self.features = f.attrs.get("features", ["fft"])
        # add found features as self.feature_name = FeatureProxy(self.h5_file, feature_name)
        self.params = Namespace()
        self._register_features(self.features, keep_open)

    def copy(self):
        return type(self)(self.h5_file, self.keep_open)

    @staticmethod
    def extract(path, **kwargs):
        raise NotImplementedError

    @classmethod
    def make(cls, db_name, files_ext='audio', items=tuple(), **kwargs):
        walker = FileWalker(files_ext, items)
        make_root_db(db_name, walker, partial(cls.extract, **kwargs))
        return cls(db_name)

    @classmethod
    def _load(cls, path, schema={}):
        """
        default extract_func for Database.build. Roughly equivalent to :
            ``{feat_name: feat.load(path) for feat_name, feat in features_dict.items()}``
        """
        out = {}
        for f_name, f in schema.items():
            # check that f has `load` and that `path` is of the right type
            if getattr(type(f), 'load', Feature.load) != Feature.load and \
                    os.path.splitext(path)[-1].strip('.') in EXTENSIONS[f.__ext__]:
                obj = f.load(path)
            # feature is derived from the output of an other
            elif hasattr(f, 'input_key') and f.input_key in out:
                # we us __call__ instead of load, the former being eq. to a transform
                obj = f(out[f.input_key][1])
            else:
                obj = None
            # obj is (data, regions)
            if isinstance(obj, tuple) and tuple(map(type, obj)) == (np.ndarray, Regions):
                out[f_name] = getattr(f, 'params', {}), *obj
            elif isinstance(obj, Regions):
                out[f_name] = getattr(f, 'params', {}), np.array([]), obj
            elif isinstance(obj, np.ndarray):
                out[f_name] = getattr(f, 'params', {}), obj, None
        return out

    @classmethod
    def create(cls, db_name, sources=tuple(), schema={}):
        """
        creates a db from the schema provided in `features_dict` and the files or root directories found in `items`

        Parameters
        ----------
        db_name : str
            the name of the file to be created
        sources : str or iterable of str
            the sources to be passed to `FileWalker`
        schema : dict
            keys (`str`) are the names of the features, values are `Feature` objects

        Returns
        -------
        db : Database
            an instance of the created db

        """
        # get the set of file extensions from the features and instantiate a walker
        exts = {f.__ext__ for f in schema.values() if getattr(f, '__ext__', False)}
        walker = FileWalker(exts, sources)
        # run the extraction job
        make_root_db(db_name, walker, partial(cls._load, schema=schema))
        # add post-build features
        db = cls(db_name)
        for f_name, f in schema.items():
            # let features the chance to update them selves confronted to their whole set
            if getattr(type(f), "post_create", Feature.post_create) != Feature.post_create:
                rv = f.after_create(db, f_name)
                if isinstance(rv, np.ndarray):
                    rv = (rv, getattr(db, f_name).regions)
                elif isinstance(rv, Regions):
                    rv = (getattr(db, f_name)[:], rv)
                elif tuple(*map(type, rv)) == (np.ndarray, Regions):
                    pass
                write_feature(db_name, f_name, getattr(f, 'params', {}),
                              *rv, files=getattr(db, f_name).files)
        db = cls(db_name)
        db.schema = schema
        return db

    @classmethod
    def make_temp(cls, roots=None, files=None, **kwargs):
        return cls.make("/tmp/%s-tmp.h5" % cls.__name__, roots, files, **kwargs)

    def _visit(self, func=print):
        with h5py.File(self.h5_file, "r") as f:
            f.visititems(func)

    def _register_features(self, names, keep_open):
        for name in names:
            setattr(self, name, FeatureProxy(self.h5_file, name, keep_open))
            setattr(self.params, name, getattr(self, name).attrs)
        return self

    def _register_dataframes(self, names):
        for name in names:
            try:
                df = pd.read_hdf(self.h5_file, key=name)
            except KeyError:
                df = pd.DataFrame()
            setattr(self, name, df)
        return self

    def __repr__(self):
        return "<Database: '%s'>" % os.path.split(self.h5_file)[-1]

    def bind(self, dataset_cls):
        """
        extends and if necessary overwrites the class of this instance with `dataset_cls`

        if `dataset_cls` contains a `__init__` method, it must take 2 arguments (self, model=None)
        and will be renamed to `prepare_dataset`.

        for example usage of this method, see `mimikit.data.DataModule.prepare_data()`

        Parameters
        ----------
        dataset_cls: type
            the interface to be added to the class of self

        Returns
        -------
        self
            the extended instance
        """
        def _bind(func, as_name=None):
            """
            Bind the function *func* to the class of *self*
            """
            if as_name is None:
                as_name = func.__name__
            setattr(type(self), as_name, func)

        for k, v in dataset_cls.__dict__.items():
            if k in ('__init__', 'prepare_dataset'):
                assert v.__code__.co_argcount == 2, \
                    "__init__ method of dataset_cls should take 2 arguments: self, model=None"
                # rename init
                _bind(v, 'prepare_dataset')
            elif k in ('__len__', '__getitem__') or k not in type(self).__dict__:
                if '__call__' in dir(v):
                    _bind(v, k)
                else:
                    # class attributes are likewise attached to type(self)
                    setattr(type(self), k, v)

        return self

    def prepare_dataset(self, model: pl.LightningModule):
        """
        placeholder for implementing what need to be done before serving data

        Parameters
        ----------
        model : pl.LightningModule
            The model that will consume this dataset.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError

    def split(self, splits):
        """
        performs random splits on self

        Parameters
        ----------
        splits: Sequence of floats or ints possibly containing None.
            The sequence of elements corresponds to the proportion (floats), the number of examples (ints) or the absence of
            train-set, validation-set, test-set, other sets... in that order.

        Returns
        -------
        splits: tuple
            the *-sets
        """
        nones = []
        if any(x is None for x in splits):
            if splits[0] is None:
                raise ValueError("the train-set's split cannot be None")
            nones = [i for i, x in zip(range(len(splits)), splits) if x is None]
            splits = [x for x in splits if x is not None]
        if all(type(x) is float for x in splits):
            splits = [x / sum(splits) for x in splits]
            N = len(self)
            # leave the last one out for now because of rounding
            as_ints = [int(N * x) for x in splits[:-1]]
            # check that the last is not zero
            if N - sum(as_ints) == 0:
                raise ValueError("the last split rounded to zero element. Please provide a greater float or consider "
                                 "passing ints.")
            as_ints += [N - sum(as_ints)]
            splits = as_ints
        sets = list(random_split(self, splits))
        if any(nones):
            sets = [None if i in nones else sets.pop(0) for i in range(len(sets + nones))]
        return tuple(sets)

    def to_tensor(self):
        for feat in self.features:
            as_tensor = self._to_tensor(getattr(self, feat))
            setattr(self, feat, as_tensor)
        return self

    def to(self, device):
        for feat in self.features:
            self._to(getattr(self, feat), device)
        return self

    @staticmethod
    def _to_tensor(obj):
        if type(obj) is torch.Tensor:
            return obj
        # converting obj[:] makes sure we get the data out of any db.feature object
        maybe_tensor = default_convert(obj[:])
        if type(maybe_tensor) is torch.Tensor:
            return maybe_tensor
        try:
            obj = torch.tensor(obj)
        except Exception as e:
            raise e
        return obj

    @staticmethod
    def _to(obj, device):
        """move any underlying tensor to some device"""
        if getattr(obj, "to", False):
            return obj.to(device)
        raise TypeError("object %s has no `to()` attribute" % str(obj))


def join(target_name, feature_names, databases):

    with h5py.File(target_name, "w") as f:
        for feat_name in feature_names:
            lengths = [getattr(db, feat_name).shape[0] for db in databases]
            dim = set([getattr(db, feat_name).shape[1] for db in databases]).pop()
            layout = h5py.VirtualLayout(shape=(sum(lengths), dim))
            offset = 0
            for i, n in enumerate(lengths):
                vsource = h5py.VirtualSource(databases[i].h5_file, feat_name, shape=(n, dim))
                layout[offset:offset + n] = vsource
                offset += n
            ds = f.create_virtual_dataset(feat_name, layout)
            ds.attrs = getattr(databases[0], feat_name).attrs



import inspect
import click
import torch
import torch.nn as nn
import dataclasses as dtc

from ..audios.features import MuLawSignal
from ..data import TBPTTSampler, Feature, AsSlice, AsFramedSlice, Input, Target
from .parts import SuperAdam, SequenceModel
from ..networks.sample_rnn import SampleRNNNetwork

from . import model, IData

__all__ = [
    'SampleRNNData',
    'SampleRNN',
    'main'
]


@dtc.dataclass
class SampleRNNData(IData):
    feature: Feature = None
    batch_size: int = 16
    chunk_len: int = 16000 * 8
    batch_seq_len: int = 512

    @classmethod
    def schema(cls, sr=22050, emphasis=0., q_levels=256):
        return {'qx': MuLawSignal(sr=sr, emphasis=emphasis, q_levels=q_levels)}

    def batch_signature(self, stage='fit'):
        batch_seq_len, frame_sizes = tuple(getattr(self, key) for key in ["batch_seq_len", "frame_sizes"])
        shifts = [frame_sizes[0] - size for size in frame_sizes]
        inputs = []
        for fs, shift in zip(frame_sizes[:-1], shifts[:-1]):
            inputs.append(
                Input('qx', AsFramedSlice(shift, batch_seq_len, frame_size=fs,
                                          as_strided=False)))
        inputs.append(
            Input('qx', AsFramedSlice(shifts[-1], batch_seq_len, frame_size=frame_sizes[-1],
                                      as_strided=True)))
        targets = Target('qx', AsSlice(shift=frame_sizes[0], length=batch_seq_len))
        if stage in ('fit', 'train', 'val'):
            return inputs, targets
        # test, predict, generate
        return Input('qx', AsSlice(0, batch_seq_len))

    @classmethod
    def dependant_hp(cls, db):
        return dict(feature=db.schema['qx'], q_levels=db.schema['qx'].q_levels)

    def loader_kwargs(self, stage, datamodule):
        ds = datamodule.datasets[stage]
        N = len(ds)
        batch_sampler = TBPTTSampler(N,
                                     self.batch_size,
                                     self.chunk_len,
                                     self.batch_seq_len)
        return dict(batch_sampler=batch_sampler)


@model
class SampleRNN(
    # db schema, batch sig
    SampleRNNData,
    # training step, generate routine & interface
    SequenceModel,
    # optimizer & scheduler
    SuperAdam,
    # nn.Module, generate_ function
    SampleRNNNetwork
):

    def setup(self, stage: str):
        SequenceModel.setup(self, stage)
        SuperAdam.setup(self, stage)

    @staticmethod
    def loss_fn(output, target):
        criterion = nn.CrossEntropyLoss(reduction="mean")
        return {"loss": criterion(output.view(-1, output.size(-1)), target.view(-1))}

    def on_train_batch_start(self, batch, batch_idx, dataloader_idx):
        if (batch_idx * self.batch_seq_len) % self.chunk_len == 0:
            self.reset_h0()

    def encode_inputs(self, inputs: torch.Tensor):
        return self.feature.encode(inputs)

    def decode_outputs(self, outputs: torch.Tensor):
        return self.feature.decode(outputs)


def make_click_command(f):
    aspec = inspect.getfullargspec(f)
    if aspec.defaults is not None:
        for arg, default in zip(reversed(aspec.args), reversed(aspec.defaults)):
            opt = click.option('--' + arg.replace("_", "-"), default=default)
            f = opt(f)
    return click.command()(f)


@make_click_command
def main(sources=['./data'], sr=16000, q_levels=256):
    import mimikit as mmk
    import os

    schema = mmk.SampleRNN.schema(sr, 0., q_levels)

    net = mmk.SampleRNN(
        feature=schema['qx'],
        q_levels=q_levels,
        frame_sizes=(16, 4, 4),
        n_rnn=2,

        max_lr=7e-4,
        betas=(.9, .91),
        div_factor=5,
    )

    print(net.hparams)

    dm = mmk.DataModule(net, "/tmp/srnn.h5",
                        sources=sources, schema=schema,
                        splits=tuple(),
                        in_mem_data=True)

    cb = mmk.GenerateCallBack(5, indices=[None] * 4,
                              n_steps=16000*10,
                              play_audios=False,
                              plot_audios=False,
                              log_audios=True,
                              log_dir=os.path.abspath('outputs/sample-rnn'),
                              temperature=torch.tensor([[.9], [.999], [1.1], [1.25]]).to('cuda'))

    trainer = mmk.get_trainer(root_dir=None,
                              max_epochs=50,
                              callbacks=[cb],
                              checkpoint_callback=False)

    trainer.fit(net, datamodule=dm)


if __name__ == '__main__':
    main()



import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
import dataclasses as dtc

from ..audios import Spectrogram
from ..data import Feature, AsSlice, Input, Target
from .parts import SuperAdam, SequenceModel, IData
from ..networks import Seq2SeqLSTM
from .parts.loss_functions import mean_L1_prop

__all__ = [
    'Seq2SeqLSTMModel'
]


@dtc.dataclass
class Seq2SeqData(IData):
    feature: Feature = None
    batch_size: int = 16

    @classmethod
    def schema(cls, sr=22050, emphasis=0., n_fft=2048, hop_length=512):
        schema = {"fft": Spectrogram(sr=sr, emphasis=emphasis,
                                     n_fft=n_fft, hop_length=hop_length,
                                     magspec=True)}
        return schema

    @classmethod
    def dependant_hp(cls, db):
        return dict(
            feature=Spectrogram(**db.fft.attrs), input_dim=db.fft.shape[-1]
        )

    def batch_signature(self, stage='fit'):
        inpt = Input('fft', AsSlice(shift=0, length=self.shift))
        trgt = Target('fft', AsSlice(shift=self.shift,
                                     length=self.shift))
        if stage in ('full', 'fit', 'train', 'val'):
            return inpt, trgt
        # test, predict, generate...
        return inpt

    def loader_kwargs(self, stage, datamodule):
        return dict(
            batch_size=self.batch_size,
            drop_last=False,
            shuffle=True
        )


class Seq2SeqLSTMModel(
    Seq2SeqData,
    SequenceModel,
    SuperAdam,
    Seq2SeqLSTM,
):

    @staticmethod
    def loss_fn(output, target):
        return {"loss": mean_L1_prop(output, target)}

    def encode_inputs(self, inputs: torch.Tensor):
        return self.feature.encode(inputs)

    def decode_outputs(self, outputs: torch.Tensor):
        return self.feature.decode(outputs)



import torch
from pytorch_lightning import LightningModule
import dataclasses as dtc
from typing import Tuple, Optional

__all__ = [
    'Adam',
    'SuperAdam'
]


class Adam(LightningModule):
    def __init__(self,
                 max_lr=1e-3,
                 betas=(.9, .9)
                 ):
        super(LightningModule, self).__init__()
        self.max_lr = max_lr
        self.betas = betas

    def configure_optimizers(self):
        self.opt = torch.optim.Adam(self.parameters(), lr=self.max_lr, betas=self.betas)
        return self.opt


class ManyOneCycleLR(torch.optim.lr_scheduler.OneCycleLR):
    def step(self, epoch=None):
        try:
            super(ManyOneCycleLR, self).step(epoch=epoch)
        except ValueError:
            self.last_epoch = 0
            self._step_count = 1
            super(ManyOneCycleLR, self).step(epoch=epoch)


@dtc.dataclass(repr=False, eq=False)
class SuperAdam(LightningModule):

    max_lr: float = 1e-3
    betas: Tuple[float, float] = (.9, .9)
    div_factor: float = 3.
    final_div_factor: float = 1.
    pct_start: float = 0.
    cycle_momentum: bool = False
    total_steps: Optional[int] = None

    def __post_init__(self):
        super(LightningModule, self).__init__()
        # those are set in setup when we know what the trainer and datamodule got
        self.steps_per_epoch = None
        self.max_epochs = None
        self.opt, self.sched = None, None

    def configure_optimizers(self):
        self.opt = torch.optim.Adam(self.parameters(), lr=self.max_lr, betas=self.betas)
        steps = dict(total_steps=self.total_steps) if self.total_steps is not None \
            else dict(steps_per_epoch=self.steps_per_epoch, epochs=self.max_epochs)
        self.sched = ManyOneCycleLR(self.opt,
                                    **steps,
                                    max_lr=self.max_lr, div_factor=self.div_factor,
                                    final_div_factor=self.final_div_factor,
                                    pct_start=self.pct_start,
                                    cycle_momentum=self.cycle_momentum
                                    )
        return [self.opt], [{"scheduler": self.sched, "interval": "step", "frequency": 1}]

    def setup(self, stage: str):
        if stage == "fit":
            self.max_epochs = self.trainer.max_epochs
            self.steps_per_epoch = len(self.train_dataloader())


class RMS(LightningModule):
    def __init__(self,
                 lr=1e-2,
                 alpha=0.99,
                 eps=1e-8,
                 weight_decay=0,
                 momentum=0,
                 centered=False
                 ):
        super(LightningModule, self).__init__()
        self.args = dict(lr=lr, alpha=alpha, eps=eps, weight_decay=weight_decay, momentum=momentum, centered=centered)

    def configure_optimizers(self):
        self.opt = torch.optim.RMSprop(self.parameters(), **self.args)
        return self.opt


import dataclasses as dtc
import torch.nn as nn
import torch

from ..abstract.features import SegmentLabels, FilesLabels
from ..audios import MuLawSignal
from ..data import Feature, Input, AsSlice, Target
from . import model
from ..networks import WNNetwork
from .parts import IData, SuperAdam, SequenceModel

__all__ = [
    'WaveNetData',
    'WaveNet'
]


@dtc.dataclass
class WaveNetData(IData):
    feature: Feature = None
    batch_size: int = 32
    batch_seq_length: int = 64

    @classmethod
    def schema(cls, sr=22050, emphasis=0., q_levels=256,
               segment_labels=False, files_labels=False):

        schema = {"qx": MuLawSignal(sr=sr, emphasis=emphasis, q_levels=q_levels)}
        if segment_labels:
            schema.update({
                'loc': SegmentLabels(input_key='qx')
            })
        if files_labels:
            schema.update({
                'glob': FilesLabels(input_key='qx')
            })
        return schema

    @classmethod
    def dependant_hp(cls, db):
        hp = {}
        if 'loc' in db.features:
            hp.update(dict(n_cin_classes=len(db.loc.regions)))
        if 'glob' in db.features:
            hp.update(dict(n_gin_classes=len(db.glob.files)))
        return dict(
            feature=db.schema['qx'], q_levels=db.schema['qx'].q_levels, **hp
        )

    def batch_signature(self, stage='fit'):
        inpt = Input('qx', AsSlice(shift=0, length=self.batch_seq_length))
        trgt = Target('qx', AsSlice(shift=self.shift,
                                    length=self.output_shape((-1, self.batch_seq_length, -1))[1]))
        # where are we conditioned?
        loc, glob = self.n_cin_classes is not None, self.n_gin_classes is not None
        if loc:
            pass
        if glob:
            pass
        if stage in ('full', 'fit', 'train', 'val'):
            return inpt, trgt
        # test, predict, generate...
        return inpt

    def loader_kwargs(self, stage, datamodule):
        return dict(
            batch_size=self.batch_size,
            drop_last=False,
            shuffle=True
        )


@model
class WaveNet(
    WaveNetData,
    SequenceModel,
    SuperAdam,
    WNNetwork,
):

    @staticmethod
    def loss_fn(output, target):
        criterion = nn.CrossEntropyLoss(reduction="mean")
        return {"loss": criterion(output.view(-1, output.size(-1)), target.view(-1))}

    def setup(self, stage: str):
        SequenceModel.setup(self, stage)
        SuperAdam.setup(self, stage)

    def encode_inputs(self, inputs: torch.Tensor):
        return self.feature.encode(inputs)

    def decode_outputs(self, outputs: torch.Tensor):
        return self.feature.decode(outputs)



import torch
import torch.nn as nn
import numpy as np

from ..models.parts import tqdm
from ..data import process_batch

__all__ = [
    'GeneratingNetwork'
]


class GeneratingNetwork(nn.Module):

    def prepare_prompt(self, prompt, n_steps, at_least_nd=2):
        def _prepare(prmpt):
            if isinstance(prmpt, np.ndarray):
                prmpt = torch.from_numpy(prmpt)
            while len(prmpt.shape) < at_least_nd:
                prmpt = prmpt.unsqueeze(0)
            prmpt = prmpt.to(self.device)
            blank_shapes = prmpt.size(0), n_steps, *prmpt.size()[2:]
            return torch.cat((prmpt, torch.zeros(*blank_shapes).to(prmpt)), dim=1)

        return process_batch(prompt, lambda x: isinstance(x, (np.ndarray, torch.Tensor)), _prepare)

    @staticmethod
    def generate_tqdm(rng):
        return tqdm(rng, desc="Generate", dynamic_ncols=True,
                    leave=False, unit="step", mininterval=0.25)



import torch
import torch.nn as nn
from dataclasses import dataclass
from typing import Optional

from .wavenet import WNNetwork, WaveNetLayer
from ..modules import homs as H, ops as Ops

__all__ = [
    'FreqNetNetwork'
]


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class FreqNetNetwork(WNNetwork):
    """
    adapts Wavenet to the frequency-domain
    """
    n_layers: tuple = (4,)
    n_input_heads: int = 1
    n_output_heads: int = 1
    input_dim: int = 256
    n_cin_classes: Optional[int] = None
    cin_dim: Optional[int] = None
    n_gin_classes: Optional[int] = None
    gin_dim: Optional[int] = None
    gate_dim: int = 128
    kernel_size: int = 2
    groups: int = 1
    accum_outputs: int = 0
    pad_input: int = 0
    gated_units: bool = True
    skip_dim: Optional[int] = None
    residuals_dim: Optional[int] = None
    reverse_dilation_order: bool = False

    def inpt_(self):
        return H.Paths(
            nn.Sequential(
                nn.Linear(self.input_dim, self.gate_dim * self.n_input_heads),
                FreqNetNetwork.ChunkSum(self.n_input_heads),
                Ops.Transpose(1, 2)
            ),
            # conditioning parameters :
            nn.Sequential(
                nn.Embedding(self.n_cin_classes, self.cin_dim), Ops.Transpose(1, 2)) if self.cin_dim else None,
            nn.Sequential(
                nn.Embedding(self.n_gin_classes, self.gin_dim), Ops.Transpose(1, 2)) if self.gin_dim else None
        )

    def outpt_(self):
        return nn.Sequential(
            Ops.Transpose(1, 2),
            nn.Linear(self.gate_dim if self.skip_dim is None else self.skip_dim, self.input_dim * self.n_output_heads),
            FreqNetNetwork.ChunkSum(self.n_output_heads),
            Ops.Abs()
        )

    @staticmethod
    def predict_(outpt, temp=None):
        return outpt

    class ChunkSum(nn.Module):

        def __init__(self, n_chunks):
            super(FreqNetNetwork.ChunkSum, self).__init__()
            self.n_chunks = n_chunks

        def forward(self, x):
            return sum(torch.chunk(x, self.n_chunks, dim=-1))


import dataclasses as dtc

import numpy as np
from ..data import Feature
from ..extract import from_recurrence_matrix


@dtc.dataclass
class SegmentLabels(Feature):
    __ext__ = 'none'

    input_key: str = 'x'
    L: int = 6
    k: int = None
    sym: bool = True
    bandwidth: float = 1.
    thresh: float = 0.2
    min_dur: int = 4

    @property
    def params(self):
        return {'input_key': self.input_key}

    @property
    def encoders(self):
        return {np.ndarray: lambda X: \
            from_recurrence_matrix(X.T, L=self.L, k=self.k, sym=self.sym,
                                   bandwidth=self.bandwidth, thresh=self.thresh,
                                   min_dur=self.min_dur)}

    def post_create(self, db, schema_key):
        return getattr(db, schema_key).regions.to_labels()


@dtc.dataclass
class FilesLabels(Feature):
    __ext__ = 'none'
    input_key: str = 'x'

    @property
    def encoders(self):
        return {np.ndarray: lambda x: np.ones((x.shape[0]))}

    def post_create(self, db, schema_key):
        return getattr(db, schema_key).files.to_labels()


import torch
import pytorch_lightning as pl
from abc import ABC
import matplotlib.pyplot as plt
import os

from .hooks import MMKHooks, LoggingHooks
from ...utils import audio

__all__ = [
    'SequenceModel',
    'GenerateCallBack'
]


class SequenceModel(MMKHooks,
                    LoggingHooks,
                    pl.LightningModule,
                    ABC):
    loss_fn = None

    def __init__(self):
        super(pl.LightningModule, self).__init__()
        MMKHooks.__init__(self)
        LoggingHooks.__init__(self)

    def training_step(self, batch, batch_idx):
        batch, target = batch
        output = self.forward(batch)
        return self.loss_fn(output, target)

    def validation_step(self, batch, batch_idx):
        batch, target = batch
        output = self.forward(batch)
        L = self.loss_fn(output, target)
        return {"val_loss" if k == "loss" else k: v for k, v in L.items()}

    def setup(self, stage: str):
        if stage == "fit" and getattr(self, "logger", None) is not None:
            self.logger.log_hyperparams(self.hparams)

    def encode_inputs(self, inputs: torch.Tensor):
        raise NotImplementedError

    def decode_outputs(self, outputs: torch.Tensor):
        raise NotImplementedError

    def before_generate(self, *args, **kwargs):
        # prepare model
        self._was_training = self.training
        self._initial_device = self.device
        self.eval()
        self.to("cuda" if torch.cuda.is_available() else "cpu")
        torch.set_grad_enabled(False)

    def after_generate(self, *args, **kwargs):
        # reset model
        self.to(self._initial_device)
        self.train() if self._was_training else None
        torch.set_grad_enabled(True)

    def generate(self, prompt, n_steps=16000,
                 encode_prompt=False, decode_outputs=False,
                 **kwargs):
        # prepare model
        self.before_generate()
        if encode_prompt:
            prompt = self.encode_inputs(prompt)
        output = self.generate_(prompt, n_steps, **kwargs)

        if decode_outputs:
            output = self.decode_outputs(output)

        self.after_generate()
        return output


class GenerateCallBack(pl.callbacks.Callback):

    def __init__(self, every_n_epochs=10, indices=3, n_steps=1000,
                 plot_audios=True, play_audios=True, log_audios=False,
                 log_dir=None,
                 **gen_kwargs):
        self.every_n_epochs = every_n_epochs
        self.indices = indices
        self.n_steps = n_steps
        self.kwargs = gen_kwargs
        self.log_audios = log_audios
        self.plot_audios = plot_audios
        self.play_audios = play_audios
        self.log_dir = log_dir

    def on_train_epoch_start(self, trainer, model):
        # TODO : avoid having to do that!
        dm = trainer.datamodule
        dm.full_ds = dm._init_ds(dm.full_ds, 'fit')

    def on_epoch_end(self, trainer: pl.Trainer, model: SequenceModel):
        if (trainer.current_epoch + 1) % self.every_n_epochs != 0:
            return
        dm = trainer.datamodule
        prompt = dm.get_prompts(self.indices)
        output = model.generate(prompt, self.n_steps, decode_outputs=True, **self.kwargs)
        sr = model.feature.params.get('sr', 22050)
        hop_length = model.feature.params.get('hop_length', 512)
        for i in range(output.size(0)):
            y = output[i].detach().cpu().numpy()
            if self.plot_audios:
                plt.figure(figsize=(20, 2))
                plt.plot(y)
                plt.show()
            if self.play_audios:
                audio(y, sr=sr, hop_length=hop_length)

        if self.log_audios:
            for i in range(output.size(0)):
                filename = "epoch=%i - prompt=%i" % (trainer.current_epoch, i)
                if self.log_dir is not None:
                    os.makedirs(self.log_dir, exist_ok=True)
                    filename = os.path.join(self.log_dir, filename)
                model.log_audio(filename, output[i].unsqueeze(0), sample_rate=sr)



import pandas as pd
import numpy as np

__all__ = [
    'Regions'
]


def ssts(item, axis=0):
    arr = np.atleast_2d(item[['start', 'stop']].values)
    slices = map(lambda a: (*[slice(0, None)] * axis, slice(a[0], a[1]),), arr)
    return slices


class Regions(pd.DataFrame):
    """
    subclass of ``pandas.DataFrame`` to construct and store sequences of slices.
    """

    @property
    def _constructor(self):
        return Regions

    @staticmethod
    def _validate(obj):
        if 'start' not in obj.columns or "stop" not in obj.columns:
            raise ValueError("Must have 'start' and 'stop' columns.")

    def slices(self, time_axis=0):
        """

        """
        return ssts(self, time_axis)

    def make_contiguous(self):
        self.reset_index(drop=True, inplace=True)
        cumdur = np.cumsum(self["duration"].values)
        self.loc[:, "start"] = np.r_[0, cumdur[:-1]] if cumdur[0] != 0 else cumdur[:-1]
        self.loc[:, "stop"] = cumdur
        return self

    @staticmethod
    def from_start_stop(starts, stops, durations):
        return Regions((starts, stops, durations), index=["start", "stop", "duration"]).T

    @staticmethod
    def from_stop(stop):
        """
        integers in `stop` correspond to the prev[stop] and next[start] values.
        `stops` must contain the last index ! and it can begin with 0, but doesn't have to...
        """
        stop = np.asarray(stop)
        if stop[0] == 0:
            starts = np.r_[0, stop[:-1]]
            stop = stop[1:]
        else:
            starts = np.r_[0, stop[:-1]]
        durations = stop - starts
        return Regions.from_start_stop(starts, stop, durations)

    @staticmethod
    def from_duration(duration):
        stops = np.cumsum(duration)
        starts = np.r_[0, stops[:-1]]
        return Regions.from_start_stop(starts, stops, duration)

    @staticmethod
    def from_data(sequence, time_axis=1):
        duration = np.array([x.shape[time_axis] for x in sequence])
        return Regions.from_duration(duration)

    def to_labels(self):
        return np.hstack([np.ones((tp.duration,), dtype=np.int) * tp.Index
                          for tp in self.itertuples()])



import math
import torch
import torch.nn as nn
from dataclasses import dataclass
from typing import Optional
from itertools import accumulate
import operator

from ..modules import homs as H, ops as Ops
from .generating_net import GeneratingNetwork

__all__ = [
    'WaveNetLayer',
    'WNNetwork'
]


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class WaveNetLayer(nn.Module):
    layer_i: int
    gate_dim: int = 128
    residuals_dim: Optional[int] = None
    skip_dim: Optional[int] = None
    kernel_size: int = 2
    groups: int = 1
    cin_dim: Optional[int] = None
    gin_dim: Optional[int] = None
    gated_units: bool = True
    pad_input: int = 1
    accum_outputs: int = -1

    stride: int = 1
    bias: bool = True

    dilation: int = 1

    shift_diff = property(
        lambda self:
        (self.kernel_size - 1) * self.dilation if self.pad_input == 0 else 0
    )
    input_padding = property(
        lambda self:
        self.pad_input * (self.kernel_size - 1) * self.dilation if self.pad_input else 0
    )
    output_padding = property(
        lambda self:
        - self.accum_outputs * self.shift_diff
    )
    receptive_field = property(
        lambda self:
        self.kernel_size * self.dilation
    )

    @property
    def conv_kwargs(self):
        return dict(kernel_size=self.kernel_size, dilation=self.dilation,
                    stride=self.stride, bias=self.bias, groups=self.groups)

    @property
    def kwargs_1x1(self):
        return dict(kernel_size=1, bias=self.bias, groups=self.groups)

    def gcu_(self):
        mod = H.AddPaths(
            # core
            nn.Conv1d(self.gate_dim, self.residuals_dim, **self.conv_kwargs),
            # conditioning parameters :
            nn.Conv1d(self.cin_dim, self.residuals_dim, **self.conv_kwargs) if self.cin_dim else None,
            nn.Conv1d(self.gin_dim, self.residuals_dim, **self.conv_kwargs) if self.gin_dim else None
        )
        return H.GatedUnit(mod) if self.gated_units else nn.Sequential(mod, nn.Tanh())

    def residuals_(self):
        return nn.Conv1d(self.residuals_dim, self.gate_dim, **self.kwargs_1x1)

    def skips_(self):
        return H.Skips(
            nn.Conv1d(self.residuals_dim, self.skip_dim, **self.kwargs_1x1)
        )

    def accumulator(self):
        return H.AddPaths(nn.Identity(), nn.Identity())

    def __post_init__(self):
        nn.Module.__init__(self)

        with_residuals = self.residuals_dim is not None
        if not with_residuals:
            self.residuals_dim = self.gate_dim

        self.gcu = self.gcu_()

        if with_residuals:
            self.residuals = self.residuals_()
        else:  # keep the signature but just pass through
            self.residuals = nn.Identity()

        if self.skip_dim is not None:
            self.skips = self.skips_()
        else:
            self.skips = lambda y, skp: skp

        if self.accum_outputs:
            self.accum = self.accumulator()
        else:
            self.accum = lambda y, x: y

    def forward(self, inputs):
        x, cin, gin, skips = inputs
        if self.pad_input == 0:
            cause = self.shift_diff if x.size(2) > self.kernel_size else self.kernel_size - 1
            slc = slice(cause, None) if self.accum_outputs <= 0 else slice(None, -cause)
            padder = nn.Identity()
        else:
            slc = slice(None)
            padder = Ops.CausalPad((0, 0, self.input_padding))
        y = self.gcu((padder(x), cin, gin))
        if skips is not None and y.size(-1) != skips.size(-1):
            skips = skips[:, :, slc]
        skips = self.skips(y, skips)
        y = self.accum(self.residuals(y), x[:, :, slc])
        return y, cin[:, :, slc] if cin is not None else cin, gin[:, :, slc] if gin is not None else gin, skips

    def output_length(self, input_length):
        if bool(self.pad_input):
            # no matter what, padding input gives the same output shape
            return input_length
        # output is gonna be less than input
        numerator = input_length - self.dilation * (self.kernel_size - 1) - 1
        return math.floor(1 + numerator / self.stride)


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class WNNetwork(GeneratingNetwork):
    n_layers: tuple = (4,)
    q_levels: int = 256
    n_cin_classes: Optional[int] = None
    cin_dim: Optional[int] = None
    n_gin_classes: Optional[int] = None
    gin_dim: Optional[int] = None
    gate_dim: int = 128
    kernel_size: int = 2
    groups: int = 1
    accum_outputs: int = 0
    pad_input: int = 0
    gated_units: bool = True
    skip_dim: Optional[int] = None
    residuals_dim: Optional[int] = None
    head_dim: Optional[int] = None
    reverse_dilation_order: bool = False

    def inpt_(self):
        return H.Paths(
            nn.Sequential(nn.Embedding(self.q_levels, self.gate_dim), Ops.Transpose(1, 2)),
            nn.Sequential(nn.Embedding(self.n_cin_classes, self.cin_dim),
                          Ops.Transpose(1, 2)) if self.cin_dim else None,
            nn.Sequential(nn.Embedding(self.n_gin_classes, self.gin_dim), Ops.Transpose(1, 2)) if self.gin_dim else None
        )

    def layers_(self):
        # allow custom sequences of kernel_sizes
        if isinstance(self.kernel_size, tuple):
            assert sum(self.n_layers) == len(self.kernel_size), "total number of layers and of kernel sizes must match"
            k_iter = iter(self.kernel_size)
            dilation_iter = iter(accumulate([1, *self.kernel_size], operator.mul))
        else:
            # reverse_dilation_order leads to the connectivity of the FFTNet
            k_iter = iter([self.kernel_size] * sum(self.n_layers))
            dilation_iter = iter([self.kernel_size**(i if not self.reverse_dilation_order else block-1-i)
                                  for block in self.n_layers for i in range(block)])

        return nn.Sequential(*[
            WaveNetLayer(i if not self.reverse_dilation_order else block - 1 - i,
                         gate_dim=self.gate_dim,
                         skip_dim=self.skip_dim,
                         residuals_dim=self.residuals_dim,
                         kernel_size=next(k_iter),
                         cin_dim=self.cin_dim,
                         gin_dim=self.gin_dim,
                         groups=self.groups,
                         gated_units=self.gated_units,
                         pad_input=self.pad_input,
                         accum_outputs=self.accum_outputs,
                         dilation=next(dilation_iter)
                         )
            for block in self.n_layers for i in range(block)
        ])

    def outpt_(self):
        # TODO : Support auxiliary classifier for conditioned networks
        return nn.Sequential(
            nn.ReLU(),
            nn.Conv1d(self.gate_dim if self.skip_dim is None else self.skip_dim,
                      self.gate_dim if self.head_dim is None else self.head_dim,
                      kernel_size=1),
            nn.ReLU(),
            nn.Conv1d(self.gate_dim if self.head_dim is None else self.head_dim,
                      self.q_levels, kernel_size=1),
            Ops.Transpose(1, 2)
        )

    def __post_init__(self):
        nn.Module.__init__(self)
        self.inpt = self.inpt_()
        self.layers = self.layers_()
        self.outpt = self.outpt_()

        rf = 0
        for i, layer in enumerate(self.layers):
            if i == (len(self.layers) - 1):
                rf += layer.receptive_field
            elif self.layers[i + 1].layer_i == 0:
                rf += layer.receptive_field - 1
        self.receptive_field = rf

        if self.pad_input == 1:
            self.shift = 1
        elif self.pad_input == -1:
            self.shift = self.receptive_field
        else:
            self.shift = sum(layer.shift_diff for layer in self.layers) + 1

    def forward(self, inputs):
        x, cin, gin = self.inpt(inputs[0],
                                inputs[1] if len(inputs) > 1 else None,
                                inputs[2] if len(inputs) == 3 else None)
        y, _, _, skips = self.layers((x, cin, gin, None))
        return self.outpt(skips if skips is not None else y)

    def output_shape(self, input_shape):
        return input_shape[0], self.all_output_lengths(input_shape[1])[-1], input_shape[-1]

    def all_output_lengths(self, input_length):
        out_length = input_length
        lengths = []
        for layer in self.layers:
            out_length = layer.output_length(out_length)
            lengths += [out_length]
        return tuple(lengths)

    def generation_slices(self):
        # input is always the last receptive field
        input_slice = slice(-self.receptive_field, None)
        if self.pad_input == 1:
            # then there's only one future time-step : the last output
            output_slice = slice(-1, None)
        else:
            # for all other cases, the first output has the shift of the whole network
            output_slice = slice(None, 1)
        return input_slice, output_slice

    @staticmethod
    def predict_(outpt, temp):
        if temp is None:
            return nn.Softmax(dim=-1)(outpt).argmax(dim=-1, keepdims=True)
        else:
            return torch.multinomial(nn.Softmax(dim=-1)(outpt / temp), 1)

    def generate_(self, prompt, n_steps, temperature=0.5, benchmark=False):
        return self.generate_slow(prompt, n_steps, temperature)

    def generate_slow(self, prompt, n_steps, temperature=0.5):

        output = self.prepare_prompt(prompt, n_steps, at_least_nd=2)
        prior_t = prompt[0].size(1)

        rf = self.receptive_field
        _, out_slc = self.generation_slices()

        for t in self.generate_tqdm(range(prior_t, prior_t + n_steps)):
            inputs = tuple(map(lambda x: x[:, t - rf:t] if x is not None else None, output))
            output[0].data[:, t:t + 1] = self.predict_(self.forward(inputs)[:, out_slc],
                                                       temperature)
        return output[0]

    def generate_fast(self, prompt, n_steps, temperature=0.5):
        # TODO : add support for conditioned networks
        output = self.prepare_prompt(prompt, n_steps, at_least_nd=2)
        prior_t = prompt.size(1) if isinstance(prompt, torch.Tensor) else prompt[0].size(1)

        inpt = output[:, prior_t - self.receptive_field:prior_t]
        z, cin, gin = self.inpt(inpt, None, None)
        qs = [(z.clone(), None)]
        # initialize queues with one full forward pass
        skips = None
        for layer in self.layers:
            z, _, _, skips = layer((z, cin, gin, skips))
            qs += [(z.clone(), skips.clone() if skips is not None else skips)]

        outpt = self.outpt(skips if skips is not None else z)[:, -1:].squeeze()
        outpt = self.predict_(outpt, temperature)
        output.data[:, prior_t:prior_t + 1] = outpt

        qs = {i: q for i, q in enumerate(qs)}

        # disable padding and dilation
        dilations = {}
        for mod in self.modules():
            if isinstance(mod, nn.Conv1d) and mod.dilation != (1,):
                dilations[mod] = mod.dilation
                mod.dilation = (1,)
        for layer in self.layers:
            layer.pad_input = 0

        # cache the indices of the inputs for each layer
        lyr_slices = {}
        for l, layer in enumerate(self.layers):
            d, k = layer.dilation, layer.kernel_size
            rf = d * k
            indices = [qs[l][0].size(2) - 1 - n for n in range(0, rf, d)][::-1]
            lyr_slices[l] = torch.tensor(indices).long().to(self.device)

        for t in self.generate_tqdm(range(prior_t + 1, prior_t + n_steps)):

            x, cin, gin = self.inpt(output[:, t - 1:t], None, None)
            q, _ = qs[0]
            q = q.roll(-1, 2)
            q[:, :, -1:] = x
            qs[0] = (q, None)

            for l, layer in enumerate(self.layers):
                z, skips = qs[l]
                zi = torch.index_select(z, 2, lyr_slices[l])
                if skips is not None:
                    # we only need one skip : the first or the last of the kernel's indices
                    i = lyr_slices[l][0 if layer.accum_outputs > 0 else -1].item()
                    skips = skips[:, :, i:i + 1]

                z, _, _, skips = layer((zi, cin, gin, skips))

                if l < len(self.layers) - 1:
                    q, skp = qs[l + 1]

                    q = q.roll(-1, 2)
                    q[:, :, -1:] = z
                    if skp is not None:
                        skp = skp.roll(-1, 2)
                        skp[:, :, -1:] = skips

                    qs[l + 1] = (q, skp)
                else:
                    y, skips = z, skips

            outpt = self.outpt(skips if skips is not None else y).squeeze()
            outpt = self.predict_(outpt, temperature)
            output.data[:, t:t + 1] = outpt

        # reset the layers' parameters
        for mod, d in dilations.items():
            mod.dilation = d
        for layer in self.layers:
            layer.pad_input = self.pad_input

        return output



import h5py
import pandas as pd
from multiprocessing import cpu_count, Pool
from concurrent.futures import ThreadPoolExecutor
import os
import warnings

from .regions import Regions

__all__ = [
    'make_root_db',
    'write_feature'
]

warnings.filterwarnings("ignore", message="PySoundFile failed.")


def _sizeof_fmt(num, suffix='b'):
    for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


# Core function

def write_feature(h5_file, feature_name, attrs, data, regions=None, files=None):
    with h5py.File(h5_file, "r+") as f:
        if feature_name in f:
            f.pop(feature_name)
        ds = f.create_dataset(name=feature_name, shape=data.shape, data=data)
        ds.attrs.update(attrs)
        f.attrs["features"] = [*f.attrs["features"], feature_name]
    if regions is not None:
        pd.DataFrame(regions).to_hdf(h5_file, feature_name + "_regions", "r+")
    if files is not None:
        pd.DataFrame(files).to_hdf(h5_file, feature_name + "_files", "r+")
    return None


def file_to_h5(abs_path, extract_func=None, output_path=None, mode="w"):
    """
    apply ``extract_func`` to ``abs_path`` and write the result in a .h5 file.

    Parameters
    ----------
    abs_path : str
        path to the file to be extracted
    extract_func : function
        the function to use for the extraction - should take exactly one argument: the path to the file to be extracted.
    output_path : str or None
        name of the created .h5 file if not None, else the name of the created .h5 file
        will be of the form : "<abs_path_without_extension>.h5"
    mode : str
        the mode to use when opening the .h5 file. default is "w".

    Returns
    -------
    created_file, infos : str, dict
        the name of the created .h5 file and a ``dict`` with the keys ``"dtype"`` and ``"shape"``
    """
    # print("making .h5 for %s" % abs_path)
    if output_path is None:
        output_path = os.path.splitext(abs_path)[0] + ".h5"
    else:
        if output_path[-3:] != ".h5":
            output_path += ".h5"
    # print("!!!", abs_path)
    rv = extract_func(abs_path)
    info = {}
    if os.path.exists(output_path) and mode == 'w':
        os.remove(output_path)
    if not os.path.exists(os.path.dirname(output_path)):
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
    try:
        f = h5py.File(output_path, mode)
    except OSError as e:
        print(output_path)
        raise e
    for name, (attrs, data, regions) in rv.items():
        ds = f.create_dataset(name=name, shape=data.shape, data=data)
        ds.attrs.update(attrs)
        info[name] = {"dtype": ds.dtype, "shape": ds.shape}
        if regions is not None:
            f.close()
            pd.DataFrame(regions).to_hdf(output_path, name + "_regions", "r+")
            f = h5py.File(output_path, "r+")
        del data
    f.attrs["features"] = list(rv.keys())
    f.flush()
    f.close()
    del rv
    return output_path, info


# Multiprocessing routine

def _make_db_for_each_file(file_walker,
                           extract_func=None,
                           destination="",
                           n_cores=cpu_count(),
                           method='mp'):
    """
    apply ``extract_func`` to the files found by ``file_walker``

    If file_walker found more than ``n_cores`` files, a multiprocessing ``Pool`` is used to speed up the process.

    Parameters
    ----------
    file_walker : iterable of str
        collection of files to be processed
    extract_func : function
        the function to apply to each file. Must take only one argument (the path to the file)
    n_cores : int, optional
        the number of cores used in the ``Pool``, default is the number of available cores on the system.

    Returns
    -------
    temp_dbs : list of tuples
        each tuple in the list is of the form ``("<created_file>.h5", dict(feature_name=dict(dtype=..., shape=...), ...))``
    """
    # add ".tmp_" prefix to the output_paths
    # args = [(file, extract_func, )
    args = [(file, extract_func, os.path.join(destination, file.strip('.').strip('/')))
            for n, file in enumerate(file_walker)]
    if len(args) > 1:
        if method != 'future':
            with Pool(min(n_cores, len(args))) as p:
                tmp_dbs_infos = p.starmap(file_to_h5, args)
        else:
            with ThreadPoolExecutor(max_workers=len(args)) as executor:
                tmp_dbs_infos = [info for info in executor.map(file_to_h5, *zip(*args))]
    else:
        tmp_dbs_infos = [file_to_h5(*arg) for arg in args]
    return tmp_dbs_infos


def _aggregate_db_infos(infos):
    """
    helper to reduce shapes and dtypes of several files and to concatenate their regions

    Parameters
    ----------
    infos : list of tuples
        see the returned value of ``_make_db_for_each_file`` for the expected type

    Returns
    -------
    ds_definitions : dict
        the keys are the name of the H5Datasets (features)
        the values are dictionaries with keys ("shape", "dtype", "regions") and corresponding values
    """
    paths = [x[0] for x in infos]
    features = set([feature for db in infos for feature in db[1].keys()])
    ds_definitions = {}
    for f in features:
        dtype = set([db[1][f]["dtype"] for db in infos])
        assert len(dtype) == 1, "aggregated features must be of a unique dtype"
        dtype = dtype.pop()
        shapes = [db[1][f]["shape"] for db in infos]
        dims = shapes[0][1:]
        if not all(shp[1:] == dims for shp in shapes[1:]):
            raise ValueError("all features should have the same dimensions but for the first axis." + \
                             f" feature {str(f)} returned different shapes")
        # collect the regions for the files (this is different from the possible segmentation regions!)
        regions = Regions.from_duration([s[0] for s in shapes])
        ds_shape = (regions.stop.values[-1], *dims)
        regions.index = paths
        ds_definitions[f] = {"shape": ds_shape, "dtype": dtype, "files_regions": regions}
    return ds_definitions


# Aggregating function

def _aggregate_dbs(target, tmp_dbs_infos, mode="w"):
    """
    copy a list of .h5 files as defined in ``tmp_dbs_infos`` into ``target``

    Parameters
    ----------
    target : str
        name for the target file
    tmp_dbs_infos : list
        see the returned value of ``_make_db_for_each_file`` for the expected type
    mode : str
        the mode with which ``target`` is opened the first time. Must be 'w' if ``target`` doesn't exist.
    """
    features_infos = _aggregate_db_infos(tmp_dbs_infos)

    # open the file & create the master datasets to host the features
    with h5py.File(target, mode) as f:
        # add the list of features
        f.attrs["features"] = list(features_infos.keys())
        for name, params in features_infos.items():
            f.create_dataset(name, shape=params["shape"], dtype=params["dtype"],
                             chunks=True, maxshape=(None, *params["shape"][1:]))

    # prepare args for copying the tmp_files into the appropriate regions
    args = []
    for feature, info in features_infos.items():
        # collect the arguments for copying each file into its regions in the master datasets/features
        args += [(source, feature, indices) for source, indices in
                 zip(info["files_regions"].index, info["files_regions"].slices(time_axis=0))]
        # "<feature>_files" will store the slices corresponding to the files for this feature
        files_regions = info["files_regions"]
        files_regions.index = files_regions.index.str.replace(".tmp_", "").str.replace(".h5", "")
        files_regions = files_regions.reset_index(drop=False)
        files_regions = files_regions.rename(columns={"index": "name"})
        pd.DataFrame(files_regions).to_hdf(target, feature + "_files", mode="r+")

    # copy the data
    intra_regions = {}
    with h5py.File(target, "r+") as trgt:
        for source, key, indices in args:
            with h5py.File(source, "r") as src:
                data = src[key][()]
                attrs = {k: v for k, v in src[key].attrs.items()}
                if key + "_regions" in src:
                    if key not in intra_regions:
                        intra_regions[key] = []
                    # concat the regions :
                    regions = pd.read_hdf(source, key=key + "_regions", mode="r")
                    regions.loc[:, ("start", "stop")] += indices[0].start
                    regions = regions.reset_index(drop=False)
                    # remove ".tmp_" from the source's name
                    regions.loc[:, "name"] = "".join(source.split(".tmp_"))
                    intra_regions[key] += [regions]
            trgt[key][indices] = data
            trgt[key].attrs.update(attrs)
    # "<feature>_regions" will store the segmenting slices for this feature
    for key, regions in intra_regions.items():
        if regions is not None:
            pd.concat(regions).reset_index(drop=True).to_hdf(target, key + "_regions", mode="r+")

    # remove the temp_dbs
    for src, _ in tmp_dbs_infos:
        if src != target:
            os.remove(src)


def make_root_db(db_name,
                 files_iterable=tuple(),
                 extract_func=None,
                 tmp_destination="/tmp/",
                 n_cores=cpu_count()
                 ):
    """
    extract and aggregate several files into a single .h5 Database

    Parameters
    ----------
    db_name : str
        the name of the db to be created
    files_iterable : iterable of str, optional
        the files to be extracted
    extract_func : function, optional
        the function to use for the extraction - should take exactly one argument.
        the default transforms the file to the stft with n_fft=2048 and hop_length=512.
    tmp_destination : str, optional
    n_cores : int, optional
        the number of cores to use to parallelize the extraction process.
        default is the number of available cores on the system.

    Returns
    -------
    db : Database
        the created db
    """
    tmp_dbs_infos = _make_db_for_each_file(files_iterable, extract_func, tmp_destination, n_cores)
    _aggregate_dbs(db_name, tmp_dbs_infos, "w")



from .freqnet import *
from .parametrized_gaussian import *
from .sample_rnn import *
from .seq2seq_lstms import *
from .tied_autoencoder import *
from .transformers import *
from .wavenet import *


__all__ = [_ for _ in dir() if not _.startswith("_")]


import torch
import torch.nn as nn

__all__ = [
    'angular_distance',
    'cosine_similarity',
    'mean_L1_prop'
]


def mean_L1_prop(output, target):
    L = nn.L1Loss(reduction="none")(output, target).sum(dim=(0, -1), keepdim=True)
    return 100 * (L / target.abs().sum(dim=(0, -1), keepdim=True)).mean()


def cosine_similarity(X, Y, eps=1e-10):
    """
    safely computes the cosine similarity between matrices X and Y.

    Shapes:
    -------
    X : (*, N, D)
    Y : (*, M, D)
    D_xy : (*, N, M)

    Notes:
    ------
    The need for this function arises from the fact that torch.nn.CosineSimilarity only computes the
    diagonal of D_xy, as in cosine_sim(output, target)
    """
    if not isinstance(eps, torch.Tensor):
        eps = torch.tensor(eps).to(X)

    dot_prod = torch.matmul(X, Y.transpose(-2, -1))
    norms = torch.norm(X, p=2, dim=-1).unsqueeze(-1) * torch.norm(Y, p=2, dim=-1).unsqueeze(-2)
    cos_theta = dot_prod / torch.maximum(norms, eps)
    return cos_theta


def angular_distance(X, Y, eps=1e-20):
    """
    angular distance is a valid distance metric based on the cosine similarity
    see https://en.wikipedia.org/wiki/Cosine_similarity#Angular_distance_and_similarity

    Shapes:
    -------
    X : (*, N, D)
    Y : (*, M, D)
    D_xy : (*, N, M)
    """
    if not isinstance(eps, torch.Tensor):
        eps = torch.tensor(eps).to(X)

    def safe_acos(x):
        # torch.acos returns nan near -1 and 1... see https://github.com/pytorch/pytorch/issues/8069
        return torch.acos(torch.clamp(x, min=-1 + eps / 2, max=1 - eps / 2))

    have_negatives = torch.any(X < 0) or torch.any(Y < 0)
    cos_theta = cosine_similarity(X, Y, eps)

    pi = torch.acos(torch.zeros(1)).item() * 2
    D_xy = (1 + int(not have_negatives)) * safe_acos(cos_theta) / pi

    return D_xy


import numpy as np
from librosa.segment import recurrence_matrix
from librosa.util import localmax
from scipy.ndimage import convolve
from ..data.regions import Regions

__all__ = [
    'from_recurrence_matrix'
]


def checker(N):
    block = np.zeros((N * 2 + 1, N * 2 + 1), dtype=np.int32)
    for k in range(-N, N + 1):
        for l in range(-N, N + 1):
            block[k + N, l + N] = np.sign(k) * np.sign(l)
    return block / abs(block).sum()


def from_recurrence_matrix(X,
                           L=6,
                           k=None,
                           sym=True,
                           bandwidth=1.,
                           thresh=0.2,
                           min_dur=4):

    R = recurrence_matrix(
        X, metric="cosine", mode="affinity",
        k=k, sym=sym, bandwidth=bandwidth, self=True)
    # intensify checker-board-like entries
    R_hat = convolve(R, checker(L),
                     mode="constant")
    # extract them along the main diagonal
    dg = np.diag(R_hat, 0)
    mx = localmax(dg * (dg > thresh)).nonzero()[0]
    # filter out maxes less than min_dur frames away of the previous max
    mx = mx * (np.diff(mx, append=R.shape[0]) >= min_dur)
    mx = mx[mx > 0]
    stops = np.r_[mx, R.shape[0]]
    return Regions.from_stop(stops)



import os
from typing import Iterable

__all__ = [
    "EXTENSIONS",
    "FileWalker"
]

AUDIO_EXTENSIONS = {"wav", "aif", "aiff", "mp3", "m4a", "mp4"}
IMAGE_EXTENSIONS = {'png', 'jpeg'}
MIDI_EXTENSIONS = {"mid"}

EXTENSIONS = dict(audio=AUDIO_EXTENSIONS,
                  img=IMAGE_EXTENSIONS,
                  midi=MIDI_EXTENSIONS,
                  none={})


class FileWalker:

    def __init__(self, files_ext, sources=None):
        """
        recursively find files from `items` whose extensions match the ones implied in `files_ext`

        Parameters
        ----------
        files_ext : str or list of str
            type(s) of files' extensions to be matched. Must be either 'audio' or 'midi'.
        sources : str or iterable of str
            a single path (string, os.Path) or an iterable of paths. Each item can either be the root of
            a directory which will be searched recursively or a single file.

        Examples
        --------
        >>> files = list(FileWalker(files_ext='midi', sources=["my-root-dir", 'piece.mid']))

        """
        if isinstance(files_ext, str):
            files_ext = [files_ext]
        if not all(ext in EXTENSIONS.keys() for ext in files_ext):
            raise ValueError("Expected all files_ext to be one of 'audio' or 'midi'.")
        self._matching_ext = {ext for file_type in files_ext for ext in EXTENSIONS[file_type]}

        generators = []

        if sources is not None and isinstance(sources, Iterable):
            if isinstance(sources, str):
                if not os.path.exists(sources):
                    raise FileNotFoundError("%s does not exist." % sources)
                if os.path.isdir(sources):
                    generators += [self.walk_root(sources)]
                else:
                    if self.is_matching_file(sources):
                        generators += [[sources]]
            else:
                for item in sources:
                    if not os.path.exists(item):
                        raise FileNotFoundError("%s does not exist." % item)
                    if os.path.isdir(item):
                        generators += [self.walk_root(item)]
                    else:
                        if self.is_matching_file(item):
                            generators += [[item]]

        self.generators = generators

    def __iter__(self):
        for generator in self.generators:
            for file in generator:
                yield file

    def walk_root(self, root):
        for directory, _, files in os.walk(root):
            for file in filter(self.is_matching_file, files):
                yield os.path.join(directory, file)

    def is_matching_file(self, filename):
        # filter out hidden files
        if os.path.split(filename.strip("/"))[-1].startswith("."):
            return False
        return os.path.splitext(filename)[-1].strip(".") in self._matching_ext



import torch.nn.functional as F
from torch import nn

__all__ = [
    'Abs',
    'Transpose',
    'CausalPad',
    'Slice',
    'Clone'
]


class Abs(nn.Module):

    def forward(self, x):
        return x.abs_()


class Transpose(nn.Module):

    def __init__(self, dim1, dim2):
        super(Transpose, self).__init__()
        self.dims = (dim1, dim2)

    def forward(self, *args):
        if len(args) > 1:
            return tuple(x.transpose(*self.dims).contiguous() if x is not None else x for x in args)
        return args[0].transpose(*self.dims).contiguous() if args[0] is not None else None


class CausalPad(nn.Module):

    @staticmethod
    def int_to_lr(i):
        return (i, 0) if i >= 0 else (0, abs(i))

    def __init__(self, pad, **kwargs):  # TODO: learn=True
        super().__init__()
        # for whatever reason torch decided dimensions should be in the reversed order...
        self.pad = tuple(i for p in reversed(pad) for i in self.int_to_lr(p))
        self.kwargs = kwargs

    def forward(self, x):
        return F.pad(x, self.pad, **self.kwargs)


class Slice(nn.Module):

    def __init__(self, slc):
        super(Slice, self).__init__()
        self.slc = slc

    def forward(self, x):
        return x[self.slc].contiguous()


class Clone(nn.Module):
    def forward(self, x):
        return x.clone()


import torch
import dataclasses as dtc

from ..abstract.features import Feature, SegmentLabels, FilesLabels
from ..audios import Spectrogram
from ..data import Input, AsSlice, Target
from .parts import SuperAdam, SequenceModel, mean_L1_prop, IData
from .model import model
from ..networks import FreqNetNetwork, WNNetwork

__all__ = [
    'FreqNetData',
    'FreqNet',
    'demo'
]


@dtc.dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class FreqNetData(IData):

    feature: Feature = None
    batch_size: int = 64
    batch_seq_length: int = 64

    @classmethod
    def schema(cls, sr=22050, emphasis=0., n_fft=2048, hop_length=512,
               segment_labels=False, files_labels=False):

        schema = {"fft": Spectrogram(sr=sr, emphasis=emphasis,
                                     n_fft=n_fft, hop_length=hop_length,
                                     magspec=True)}
        if segment_labels:
            schema.update({
                'loc': SegmentLabels(input_key='fft')
            })
        if files_labels:
            schema.update({
                'glob': FilesLabels(input_key='fft')
            })
        return schema

    @classmethod
    def dependant_hp(cls, db):
        hp = {}
        if 'loc' in db.features:
            hp.update(dict(n_cin_classes=len(db.loc.regions)))
        if 'glob' in db.features:
            hp.update(dict(n_gin_classes=len(db.glob.files)))
        return dict(
            feature=Spectrogram(**db.fft.attrs), input_dim=db.fft.shape[-1], **hp
        )

    def batch_signature(self, stage='fit'):
        inpt = [Input('fft', AsSlice(shift=0, length=self.batch_seq_length))]
        trgt = Target('fft', AsSlice(shift=self.shift,
                                     length=self.output_shape((-1, self.batch_seq_length, -1))[1]))
        # where are we conditioned?
        loc, glob = self.n_cin_classes is not None, self.n_gin_classes is not None
        if loc:
            inpt += [Input('loc', AsSlice(shift=0, length=self.batch_seq_length))]
        if glob:
            inpt += [Input('glob', AsSlice(shift=0, length=self.batch_seq_length))]
        if stage in ('full', 'fit', 'train', 'val'):
            return inpt, trgt
        # test, predict, generate...
        return inpt

    def loader_kwargs(self, stage, datamodule):
        return dict(
            batch_size=self.batch_size,
            drop_last=False,
            shuffle=True
        )


@model
class FreqNet(
    # data configuration :
    FreqNetData,
    # optimizer :
    SuperAdam,
    # training step, generate routine & interface :
    SequenceModel,
    # overrides wavenet's inputs/outputs modules :
    FreqNetNetwork,
    # we inherit the networks __init__ and all generate_* from wavenet
    WNNetwork
):

    @staticmethod
    def loss_fn(output, target):
        return {"loss": mean_L1_prop(output, target)}

    def setup(self, stage: str):
        SequenceModel.setup(self, stage)
        SuperAdam.setup(self, stage)

    def encode_inputs(self, inputs: torch.Tensor):
        return self.feature.encode(inputs)

    def decode_outputs(self, outputs: torch.Tensor):
        return self.feature.decode(outputs)


# this script is a base template for programmatically generating
# demo notebooks
def demo():
    """### import and arguments"""
    import mimikit as mmk

    # DATA

    # list of files or directories to use as data
    sources = ['./data']
    # audio sample rate
    sr = 22050
    # the size of the stft
    n_fft = 2048
    # hop_length of the
    hop_length = n_fft // 4

    # NETWORK

    # the number of layers determines 'how much past' is used to predict the next future step
    # here you can make blocks of layers by specifying a tuple of integers, e.g. (2, 3, 2)
    n_layers = (3, )
    # how many parameters pro layer (must be divisible by 2)
    gate_dim = 1024
    # this multiplies the number of parameters used in the input & output layers.
    n_heads = 4

    # OPTIMIZATION

    # how many epochs should we train for
    max_epochs = 50
    # how many examples are used pro training steps
    batch_size = 16
    # how long are the examples in each training step
    # must be bigger than the network's receptive field
    batch_seq_length = sum(2**n for n in n_layers) * 2
    # the learning rate
    max_lr = 5e-4
    # betas control how fast the network changes its 'learning course'.
    # generally, betas should be close but smaller than 1. and be balanced with the batch_size :
    # the smaller the batch, the higher the betas 'could be'.
    betas = (0.9, 0.93)

    # MONITORING

    # how often should the network generate during training
    every_n_epochs = 2
    # how many examples from random prompts should be generated
    n_examples = 3
    # how many steps (1 step = 1 fft frame) should be generated
    n_steps = 1000

    # make sure batches are long enough
    rf = sum(2**n for n in n_layers)
    assert batch_seq_length > rf, f"batch_seq_length ({batch_seq_length}) needs to be greater than the receptive field ({rf})"
    print("arguments are ok!")

    """### create the data"""
    schema = mmk.FreqNet.schema(sr, n_fft=n_fft, hop_length=hop_length,
                                segment_labels=False, files_labels=False)

    db_path = 'freqnet-demo.h5'
    print("collecting data...")
    db = mmk.Database.create(db_path, sources, schema)
    print("successfully created the db.")

    """### create network and train"""
    net = mmk.FreqNet(
        **mmk.FreqNet.dependant_hp(db),
        n_layers=n_layers,
        gate_dim=gate_dim,
        groups=2,
        n_input_heads=n_heads,
        n_output_heads=n_heads,
        batch_size=batch_size,
        batch_seq_length=batch_seq_length,
        max_lr=max_lr,
        div_factor=5,
        betas=betas,

    )
    print(net.hparams)

    dm = mmk.DataModule(net, db, splits=tuple())

    cb = mmk.GenerateCallBack(every_n_epochs,
                              indices=[None]*n_examples,
                              n_steps=n_steps,
                              play_audios=True,
                              plot_audios=True)

    trainer = mmk.get_trainer(root_dir=None,
                              max_epochs=max_epochs,
                              callbacks=[cb],
                              checkpoint_callback=False)
    print("here we go!")
    trainer.fit(net, datamodule=dm)

    """----------------------------"""


if __name__ == '__main__':
    demo()



import dataclasses as dtc
import os
import pytorch_lightning as pl
import torch
import numpy as np
from numpy.lib.stride_tricks import as_strided as np_as_strided
from torch.utils.data import Dataset, DataLoader
from typing import Iterable, Optional, Callable
import re
from random import randint
from torch._six import container_abcs, string_classes

from . import Database

__all__ = [
    'AsSlice',
    'AsFramedSlice',
    'Input',
    'Target',
    'process_batch',
    'DefaultDataset',
    'DataModule'
]


@dtc.dataclass
class Getter:
    """
    base class for implementing data getter

    Parameters
    ----------

    Attributes
    ----------
    n : int or None
        the length of the underlying data
    """
    n: Optional[int] = dtc.field(default=None, init=False)

    def __call__(self, feat_data, item):
        """
        apply this instance's logic to get data from ``feat_data`` for a given ``item``

        Parameters
        ----------
        feat_data: [np.ndarray, torch.Tensor, mimikit.FeatureProxy]

        item: int
            the index emitted from a sampler

        Returns
        -------
        data: Any
            the examples corresponding to this item
        """
        return feat_data[item]

    def __len__(self):
        return self.n


@dtc.dataclass()
class AsSlice(Getter):
    """
    maps an ``item`` to a slice of data

    Parameters
    ----------
    shift : int
        the slice will start at the index `item + shift`
    length : int
        the length of the slice
    stride : int
        sub-sampling factor. Every `stride` datapoints `item` increases of `1`

    Examples
    --------

    .. testcode::

       import mimikit as mmk

       slicer = mmk.AsSlice(shift=2, length=3)
       data, item = list(range(10)), 2

       # now use it like a function :
       sliced = slicer(data, item)

       print(sliced)

    will output:

    .. testoutput::

       [4, 5, 6]
    """
    shift: int = 0
    length: int = 1
    stride: int = 1

    def __call__(self, feat_data, item):
        i = item * self.stride
        return feat_data[slice(i + self.shift, i + self.shift + self.length)]

    def __len__(self):
        return (self.n - (self.shift + self.length) + 1) // self.stride


@dtc.dataclass
class AsFramedSlice(AsSlice):
    frame_size: int = 1
    as_strided: bool = False

    def __call__(self, feat_data, item):
        sliced = super(AsFramedSlice, self).__call__(feat_data, item)
        if self.as_strided:
            if isinstance(sliced, np.ndarray):
                itemsize = sliced.dtype.itemsize
                as_strided = lambda arr: np_as_strided(arr,
                                                       shape=(self.length, self.frame_size),
                                                       strides=(itemsize, itemsize))
            else:
                as_strided = lambda tensor: torch.as_strided(tensor,
                                                             size=(self.length, self.frame_size),
                                                             stride=(1, 1))

            with torch.no_grad():
                return as_strided(sliced)
        else:
            return sliced.reshape(-1, self.frame_size)


@dtc.dataclass
class Input:
    db_key: str = ''
    getter: Getter = Getter()
    transform: Callable = lambda x: x

    def __len__(self):
        return len(self.getter)


class Target(Input):
    """exactly equivalent to Input, just makes code simpler to read."""
    pass


np_str_obj_array_pattern = re.compile(r'[SaUO]')


def process_batch(batch, test=lambda x: False, func=lambda x: x):
    """
    recursively apply func to the elements of data if test(element) is True.
    This is used in DefaultDataset to process elements (Input or Target) packed in tuples, list, dict etc...
    """
    elem_type = type(batch)
    if test(batch):
        return func(batch)
    elif isinstance(batch, container_abcs.Mapping):
        return {key: process_batch(batch[key], test, func) for key in batch}
    elif isinstance(batch, tuple) and hasattr(batch, '_fields'):  # namedtuple
        return elem_type(*(process_batch(d, test, func) for d in batch))
    elif isinstance(batch, container_abcs.Sequence) and not isinstance(batch, string_classes):
        return [process_batch(d, test, func) for d in batch]
    else:
        return batch


def _is_batchitem(obj):
    return isinstance(obj, (Input, Target))


class DefaultDataset(Dataset):

    def prepare_dataset(self, batch=tuple()):
        super(Dataset, self).__init__()

        # pass the lengths of the db features to the getters
        def cache_lengths(feat):
            if feat.getter.n is None:
                setattr(feat.getter, 'n', len(getattr(self, feat.db_key)))
            return feat

        self.batch = process_batch(batch, _is_batchitem, cache_lengths)

        # get the minimum length of all batchitems
        self.N = float('inf')

        def set_n_to_min(feat):
            self.N = min(len(feat), self.N)
            return feat

        process_batch(self.batch, _is_batchitem, set_n_to_min)

        return self

    def __getitem__(self, item):
        def get_data(feat):
            return feat.transform(feat.getter(getattr(self, feat.db_key), item))

        return process_batch(self.batch, _is_batchitem, get_data)

    def __len__(self):
        return self.N


def _implements_mapstyle(obj):
    if isinstance(obj, type):
        getitem, flen = getattr(obj, '__getitem__', False), getattr(obj, '__len__', False)
    else:
        getitem, flen = getattr(obj.__class__, '__getitem__', False), getattr(obj.__class__, '__len__', False)

    return getitem not in (Database.__getitem__, Dataset.__getitem__) and flen is not Database.__len__


@dtc.dataclass
class DataModule(pl.LightningDataModule):
    """
    encapsulate the logic for creating, initializing and serving databases & datasets
    """

    model: pl.LightningModule = None
    db: [Database, Dataset, str] = None
    keep_open: bool = True
    sources: [str, Iterable[str]] = ''
    schema: dict = dtc.field(default_factory=dict)
    dataset_cls: type = None
    in_mem_data: bool = True
    splits: tuple = tuple()
    loader_kwargs: dict = dtc.field(default_factory=dict)

    def __post_init__(self):
        super(DataModule, self).__init__()
        self.has_split = False
        self.full_ds = None
        self.datasets = {}

    def prepare_data(self, *args, **kwargs):
        """
        creates a db if necessary and possible
        """
        # get a db object (create it if necessary)
        if isinstance(self.db, (str, os.PathLike)):
            if not os.path.exists(self.db):
                # user must provide sources and schema to build db
                if not self.sources and self.schema:
                    raise ValueError("You need to provide sources and a schema in order to create a db")
                Database.create(self.db, self.sources, self.schema)

    def _instantiate_db(self):
        if isinstance(self.db, (str, os.PathLike)):
            self.db = Database(self.db, keep_open=self.keep_open)
        elif isinstance(self.db, (Database, Dataset)):
            pass
        else:
            raise TypeError("Expected `db` to be of type str, Database or Dataset. Got " + str(type(self.db)))

    def _init_ds(self, ds, stage):
        # upgrade db to a Dataset if it isn't one already
        if _implements_mapstyle(ds):
            pass
        # user provided Dataset class
        elif self.dataset_cls is not None:
            ds = ds.bind(self.dataset_cls)
        # model has a batch signature
        elif getattr(self.model, 'batch_signature', False):
            ds = ds.bind(DefaultDataset)
        else:
            raise RuntimeError("couldn't instantiate a Dataset with the provided arguments")
        if hasattr(ds, 'prepare_dataset') and hasattr(self.model, 'batch_signature'):
            ds = ds.prepare_dataset(self.model.batch_signature(stage))
        return ds

    def _split(self):
        if not self.splits:
            self.datasets['full'] = self.full_ds
            self.datasets['fit'] = self.full_ds
        else:
            sets = self.full_ds.split(self.splits)
            # store the sets as attr
            for ds, stage in zip(sets, ["fit", "val", "test"]):
                if ds is not None:
                    self.datasets[stage] = ds
        setattr(self, 'has_split', True)
        return self

    def _move_to_mem(self, ds):
        if isinstance(ds, Database):
            if self.in_mem_data and torch.cuda.is_available():
                ds.to_tensor()
                ds.to("cuda")
        return ds

    def setup(self, stage=None):
        if not self.has_prepared_data:
            self.prepare_data()
        if not isinstance(self.db, (Database, Dataset)):
            self._instantiate_db()
        if self.full_ds is None:
            self.full_ds = self._init_ds(self.db, 'fit')
            self.full_ds = self._move_to_mem(self.full_ds)
        if not self.has_split:
            self._split()
        # update loader kwargs
        if hasattr(self.model, 'loader_kwargs'):
            self.loader_kwargs = self.model.loader_kwargs(stage, self)

    def full_dataloader(self):
        self.setup()
        return DataLoader(self.full_ds, **self.loader_kwargs)

    def train_dataloader(self):
        if not self.has_setup_fit:
            self.setup("fit")
        return DataLoader(self.datasets['fit'], **self.loader_kwargs)

    def val_dataloader(self):
        if 'val' not in self.datasets:
            return None
        if not self.has_setup_fit:
            self.setup("val")
        kwargs = self.loader_kwargs.copy()
        return DataLoader(self.datasets['val'], **kwargs)

    def test_dataloader(self):
        if 'test' not in self.datasets:
            return None
        if not self.has_setup_test:
            self.setup("test")
        kwargs = self.loader_kwargs.copy()
        return DataLoader(self.datasets['test'], **kwargs)

    def get_prompts(self, indices=tuple()):
        ds = self.full_ds
        if hasattr(ds, 'prepare_dataset') and hasattr(self.model, 'batch_signature'):
            ds = ds.prepare_dataset(self.model.batch_signature('test'))
        N = len(ds)
        data = [ds[randint(0, N) if ix is None else ix] for ix in indices]
        return torch.utils.data.dataloader.default_collate(data)



import librosa
import numpy as np
from librosa.display import specshow
import IPython.display as ipd
import matplotlib.pyplot as plt

__all__ = [
    'audio',
    'show'
]

HOP_LENGTH, SR = 512, 22050

# Conversion

a2db = lambda S: librosa.amplitude_to_db(abs(S), ref=S.max())


# Debugging utils

def to_db(S):
    if S.dtype == np.complex64:
        S_hat = a2db(abs(S)) + 40
    elif S.min() >= 0 and S.dtype in (np.float, np.float32, np.float64, np.float_):
        S_hat = a2db(S) + 40
    else:
        S_hat = a2db(S) + 40
    return S_hat


def signal(S, hop_length=HOP_LENGTH):
    if S.dtype in (np.complex64, np.complex128):
        return librosa.istft(S, hop_length=hop_length)
    else:
        return librosa.griffinlim(S, hop_length=hop_length, n_iter=32)


def audio(S, hop_length=HOP_LENGTH, sr=SR):
    if len(S.shape) > 1:
        y = signal(S, hop_length)
        if y.size > 0:
            return ipd.display(ipd.Audio(y, rate=sr))
        else:
            return ipd.display(ipd.Audio(np.zeros(hop_length*2), rate=sr))
    else:
        return ipd.display(ipd.Audio(S, rate=sr))


def show(S, figsize=(), db_scale=True, title="", **kwargs):
    S_hat = to_db(S) if db_scale else S
    if figsize:
        plt.figure(figsize=figsize)
    if "x_axis" not in kwargs:
        kwargs["x_axis"] = "frames"
    if "y_axis" not in kwargs:
        kwargs["y_axis"] = "frames"
    ax = specshow(S_hat, sr=SR, **kwargs)
    plt.colorbar()
    plt.tight_layout()
    plt.title(title)
    return ax




__version__ = '0.2.1'

from . import audios
from . import connectors
from . import data
from . import extract
from . import midi
from . import models
from . import modules
from . import networks

from .audios import *
from .connectors import *
from .data import *
from .extract import *
from .midi import *
from .models.parts import *
from .models import *
from .modules import *
from .networks import *

from .file_walker import *
from .get_trainer import *
from .utils import *





from neptune import Session
from neptune.experiments import Experiment
from zipfile import ZipFile
import os
from typing import Tuple
from getpass import getpass
import shutil

try:
    from pytorch_lightning import LightningModule
except ModuleNotFoundError:
    # we only need lightning for annotation purposes
    # and setup.py allows for a "no-torch" install that contains this module
    # so we simulate the import if the module wasn't found
    class pytorch_lightning:
        LightningModule = object()

from mimikit.data import Database

__all__ = [
    'NeptuneConnector'
]


class NeptuneConnector:
    NEPTUNE_TOKEN_KEY = "NEPTUNE_API_TOKEN"

    @property
    def api_token(self):
        """
        either finds neptune's API token in os.environ[self.NEPTUNE_TOKEN_KEY]
        or prompt the user to paste it and stores it in the environment under self.NEPTUNE_TOKEN_KEY

        Returns
        -------
        api_token : str

        """
        api_token = os.environ.get(self.NEPTUNE_TOKEN_KEY, None)
        if api_token is None:
            api_token = getpass("Couldn't find your api token in this system's environment. "
                                "Please paste your API token from neptune here: ")
            os.environ[self.NEPTUNE_TOKEN_KEY] = api_token
        return api_token

    @property
    def session(self):
        """
        neptune.Session initialized with a found/prompted token

        Returns
        -------
        session : neptune.Session

        """
        if self._session is None:
            self._session = Session.with_default_backend(api_token=self.api_token)
        return self._session

    def __init__(self,
                 user: str = None,
                 setup: dict = None):
        """
        manages the integration of ``neptune`` into ``mimikit``.

        In order to connect to neptune, the ``NeptuneConnector`` probes the system's environment for an API token, and
        if none were found, it prompts the user for it and subsequently stores it in the environment.
        You can configure the name of the environment variable through the static attribute
        ``NeptuneConnector.NEPTUNE_TOKEN_KEY`` which is, by default, set to "NEPTUNE_API_TOKEN".

        Each instance of the ``NeptuneConnector`` type holds a dictionary in it's ``setup`` attribute where the keys
        are user-defined and the values are expected to be paths on a neptune-user's graph.
        Specifically, ``"<user>"`` being expected in the ``user`` argument of the ``NeptuneConnector`` constructor,
        the values in ``setup`` can be of two forms :
            1.  ``"<project>"``
            2. ``"<project>/<experiment-id>"``
        thus allowing the ``NeptuneConnector`` to
            1. create new experiments in ``"<user>/<project>"``
            2. retrieve data from specific experiments @ ``"<user>/<project>/<experiment-id>"``

        All methods of the ``NeptuneConnector`` take a ``setup_key`` argument to specify "where" the method should be
        executed.

        Parameters
        ----------
        user : str
            the neptune username
        setup : dict
            a dictionary where the values are paths in the neptune user's graph.
            see description above for details.

        """
        self.user = user
        self.setup = setup
        self._session = None
        # initialize the token right-away
        _ = self.api_token

    def path(self, setup_key: str, split: bool = False):
        """
        returns the full neptune path corresponding to ``setup_key``.

        Parameters
        ----------
        setup_key : str
            the key to look up in self.setup
        split : bool, optional
            whether to split the path. If there's no exp-id at the end of the setup_value, None is appended

        Returns
        -------
        path : str or list of str depending whether or not `split=True`

        """
        if setup_key not in self.setup:
            raise KeyError("%s not in found in setup" % setup_key)
        p = os.path.join(self.user, self.setup[setup_key])
        if split:
            p = p.split("/")
            if len(p) == 2:  # i.e. no experiment ID
                p += [None]
        return p

    def get_experiment(self, setup_key: str):
        """
        get the experiment object for setup_key

        Parameters
        ----------
        setup_key : str
            the key to look up in self.setup

        Returns
        -------
        experiment : neptune.Experiment
            the experiment object

        Raises
        ------
        ValueError if the value for the setup_key doesn't contain an experiment-id at its end
        """
        namespace, project, exp_id = self.path(setup_key, split=True)
        if exp_id is None:
            raise ValueError("'%s' has no experiment-id at the end" % self.path(setup_key))
        project = self.session.get_project(namespace + "/" + project)
        exps = project.get_experiments(id=exp_id)
        return None if not any(exps) else exps[0]

    def create_experiment(self, setup_key: str, params: dict, **kwargs):
        """
        creates a new experiment in the project of the setup_key

        Parameters
        ----------
        setup_key : str
            the key to look up in self.setup where the experiment will be created
        params : dict
            the params (hyper-parameters) of the experiment
        kwargs
            optional arguments for the native `create_experiment()` method from neptune.
            see https://docs.neptune.ai/api-reference/neptune/index.html#neptune.create_experiment

        Returns
        -------
        experiment : neptune.Experiment
            the created experiment

        """
        user, project, exp_id = self.path(setup_key, split=True)
        project_name = os.path.join(self.user, project)
        project = self.session.get_project(project_name)
        return project.create_experiment(params=params, **kwargs)

    def download_experiment(self, setup_key: str, destination: str = "./", artifacts: str = "/"):
        """
        downloads all artifacts lying at the root of an experiment into a folder named after the experiment-id

        Parameters
        ----------
        setup_key : str
            the key to look up in self.setup
        destination : str, optional
            a directory where to create the folder `"<exp-id>/"` in which the data will be unzipped.
        artifacts : str, optional
            an optional path in the artifacts to download, default is "/" i.e. all artifacts. Has to be a directory.

        Returns
        -------
        experiment : neptune.Experiment
            the neptune object for the downloaded experiment

        """
        namespace, project, exp_id = self.path(setup_key, split=True)
        project_name = namespace + "/" + project
        model_project = self.session.get_project(project_name)
        exp = model_project.get_experiments(id=exp_id)[0]
        destination = os.path.join(destination, exp_id)
        exp.download_artifacts(artifacts, destination)
        artifacts_name = "output" if artifacts == "/" else os.path.split(artifacts.strip("/"))[-1]
        with ZipFile(os.path.join(destination, artifacts_name + ".zip")) as f:
            f.extractall(destination)
        if artifacts == "/":
            for subdir in os.listdir(os.path.join(destination, artifacts_name)):
                shutil.move(src=os.path.join(destination, artifacts_name, subdir),
                            dst=os.path.join(destination, subdir))
        os.remove(os.path.join(destination, artifacts_name + ".zip"))
        if artifacts == "/":
            shutil.rmtree(os.path.join(destination, artifacts_name))
        return exp

    def get_project(self, setup_key: str):
        """
        get the project object for setup_key

        Parameters
        ----------
        setup_key : str
            the key to look up in self.setup

        Returns
        -------
        project : neptune.Project
            the project object

        """
        user, project, _ = self.path(setup_key, split=True)
        project = self.session.get_project(os.path.join(self.user, project))
        return project

    def upload_database(self, setup_key: str, db: Database):
        """
        upload .h5 file to a neptune Experiment.

        If the value of the setup_key doesn't contain an experiment-ID, one is created.
        If it does, the .h5 file is added to the top-level of the artifacts of this experiment.

        Parameters
        ----------
        setup_key : str
            the key in the setup you want to upload to
        db : Database
            the object pointing to the file you want to upload

        Returns
        -------
        `1` if the upload was successful
        """
        _, _, exp_id = self.path(setup_key, split=True)
        if exp_id is None:
            feature = [name for name in db.features if "label" not in name][0]
            feat_prox = getattr(db, feature)
            params = {"name": db.h5_file,
                      "feature_name": feature,
                      "shape": feat_prox.shape,
                      }
            params.update(feat_prox.attrs)
            exp = self.create_experiment(setup_key, params=params)
        else:
            exp = self.get_experiment(setup_key)
        exp.log_artifact(db.h5_file)
        exp.stop() if exp_id is None else None
        return 1

    def download_database(self, setup_key: str, path_to_artifact: str, destination="./"):
        """
        downloads a database (.h5 artifact) from the experiment corresponding to a setup_key

        Parameters
        ----------
        setup_key : str
            the key where to find the experiment in self.setup
        path_to_artifact :
            the path to desired artifact within the neptune's experiment
        destination : str, optional
            the path on your file-system where to put the artifact. default is "./"

        Returns
        -------
        db : Database
            an Database instance of the downloaded .h5 file
        """
        namespace, project, exp_id = self.path(setup_key, split=True)
        project_name = namespace + "/" + project
        data_project = self.session.get_project(project_name)
        exp = data_project.get_experiments(id=exp_id)[0]
        exp.download_artifact(path_to_artifact, destination)
        return Database(os.path.join(destination, os.path.split(path_to_artifact)[-1]))

    def upload_model(self,
                     setup_key: str,
                     model: [str, LightningModule],
                     artifacts: Tuple[str] = ("states", "logs", "audios")):
        """
        upload the sub-directories of a model to neptune at the specified setup_key

        Parameters
        ----------
        setup_key : str
            the key in self.setup where the data should be uploaded.
        model : str or LightningModule
            the model (``LightningModule``) to be uploaded or a root directory (``str``) for the ``artifacts``.
        artifacts : tuple of str, optional
            tuple containing the names of the sub-directories to be uploaded.
            by default `upload_model` uploads the sub-directories "states", "logs" and "audios".
            each uploaded artifact will sit at the top-level of the experiment

        Returns
        -------
        `1` if the upload was successful

        Raises
        ------
        ValueError if `model` is not a string and has neither a `trainer` nor a `_loaded_from_checkpoint` attribute
        """
        if isinstance(model, str):
            root_dir = model
            experiment = None
        else:
            if getattr(model, "trainer", None) is not None:
                root_dir = model.trainer.default_root_dir
            elif getattr(model, "_loaded_checkpoint", False):
                root_dir = os.path.split(os.path.dirname(model._loaded_checkpoint))[0]
            else:
                raise ValueError("Expected to either find a 'trainer' or a '_loaded_from_checkpoint' attribute in the"
                                 " model. Found neither.")
            experiment = [exp for exp in model.logger.experiment if isinstance(exp, Experiment)]
            experiment = None if not any(experiment) else experiment[0]
        if experiment is None:
            _, _, exp_id = self.path(setup_key, split=True)
            if exp_id is None:
                experiment = self.create_experiment(setup_key, {})
            else:
                experiment = self.get_experiment(setup_key)
        for directory in os.listdir(root_dir):
            if directory in artifacts:
                artifact = os.path.join(root_dir, directory)
                experiment.log_artifact(artifact, directory)
                print("successfully uploaded", artifact, "to", experiment.id)
        return 1



from .create import *
from .database import *
from .datamodule import *
from .feature import *
from .regions import *
from .tbptt_sampler import *


__all__ = [_ for _ in dir() if not _.startswith("_")]


import torch
import torch.nn as nn
from typing import Optional
from dataclasses import dataclass

from .generating_net import GeneratingNetwork

__all__ = [
    "SampleRNNTier",
    "SampleRNNNetwork"
]


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class SampleRNNTier(nn.Module):
    tier_index: int
    frame_size: int
    dim: int
    up_sampling: int = 1
    n_rnn: int = 2
    q_levels: int = 256
    embedding_dim: Optional[int] = None
    mlp_dim: Optional[int] = None

    is_bottom = property(lambda self: self.up_sampling == 1)

    def embeddings_(self):
        if self.embedding_dim is not None:
            return nn.Embedding(self.q_levels, self.embedding_dim)
        return None

    def input_proj_(self):

        if not self.is_bottom:  # top & middle tiers
            return nn.Sequential(nn.Linear(self.frame_size, self.dim), )

        else:  # bottom tier
            class BottomProjector(nn.Module):
                def __init__(self, emb_dim, out_dim, frame_size):
                    super(BottomProjector, self).__init__()
                    self.cnv = nn.Conv1d(emb_dim, out_dim, kernel_size=frame_size)

                def forward(self, hx):
                    """ hx : (B x T x FS x E) """
                    B, T, FS, E = hx.size()
                    hx = self.cnv(hx.view(B * T, FS, E).transpose(1, 2).contiguous())
                    # now hx : (B*T, DIM, 1)
                    return hx.squeeze().reshape(B, T, -1)

            return BottomProjector(self.embedding_dim, self.dim, self.frame_size)

    def rnn_(self):
        # no rnn for bottom tier
        if self.is_bottom:
            return None
        # we do not learn hidden for practical reasons : storing them as params in the state dict
        # would force the network to always generate with the train batch_size or to implement some logic
        # to handle this shape change. We choose the simpler solution : hidden are initialized to zeros
        # whenever we need fresh ones.
        self.h0, self.c0 = None, None
        return nn.LSTM(self.dim, self.dim, self.n_rnn, batch_first=True)

    def up_sampling_net_(self):
        # no up sampling for bottom tier
        if self.is_bottom:
            return None

        class TimeUpscalerLinear(nn.Module):

            def __init__(self, in_dim, out_dim, up_sampling, **kwargs):
                super(TimeUpscalerLinear, self).__init__()
                self.up_sampling = up_sampling
                self.out_dim = out_dim
                self.fc = nn.Linear(in_dim, out_dim * up_sampling, **kwargs)

            def forward(self, x):
                B, T, _ = x.size()
                return self.fc(x).reshape(B, T * self.up_sampling, self.out_dim)

        class ConvUpscaler(nn.Module):
            def __init__(self, dim, up_sampling):
                super(ConvUpscaler, self).__init__()
                self.cv = nn.ConvTranspose1d(dim, dim, kernel_size=up_sampling, stride=up_sampling)

            def forward(self, x):
                return self.cv(x.transpose(1, 2).contiguous()).transpose(1, 2).contiguous()

        return TimeUpscalerLinear(self.dim, self.dim, self.up_sampling)

    def mlp_(self):
        if self.mlp_dim is None:
            return None
        return nn.Sequential(
            nn.Linear(self.dim, self.mlp_dim), nn.ReLU(),
            nn.Linear(self.mlp_dim, self.mlp_dim), nn.ReLU(),
            nn.Linear(self.mlp_dim, self.q_levels if self.is_bottom else self.dim),
        )

    def __post_init__(self):
        nn.Module.__init__(self)
        self.embeddings = self.embeddings_()
        self.inpt_proj = self.input_proj_()
        self.rnn = self.rnn_()
        self.up_net = self.up_sampling_net_()
        self.mlp = self.mlp_()

    def reset_hidden(self, batch_size, device):
        self.h0 = torch.zeros(self.n_rnn, batch_size, self.dim).to(device)
        self.c0 = torch.zeros(self.n_rnn, batch_size, self.dim).to(device)
        return self.h0, self.c0

    def linearize(self, q_samples):
        """ maps input samples (0 <= qx < 256) to floats (-2. <= x < 2.) """
        return ((q_samples.float() / self.q_levels) - .5) * 4

    def forward(self, input_samples, prev_tier_output=None, hidden=None):
        if self.embeddings is None:
            x = self.linearize(input_samples)
        else:
            x = self.embeddings(input_samples)

        if self.inpt_proj is not None:
            p = self.inpt_proj(x)
            if prev_tier_output is not None:
                x = p + prev_tier_output
            else:
                x = p

        if self.rnn is not None:
            if hidden is None or x.size(0) != hidden[0].size(1):
                hidden = self.reset_hidden(x.size(0), x.device)
            else:
                # TRUNCATED back propagation through time == detach()!
                hidden = tuple(h.detach() for h in hidden)
            x, hidden = self.rnn(x, hidden)

        if self.up_net is not None:
            x = self.up_net(x)

        if self.mlp is not None:
            x = self.mlp(x)
        return x, hidden


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class SampleRNNNetwork(GeneratingNetwork, nn.Module):

    frame_sizes: tuple = (16, 8, 8)  # from top to bottom!
    dim: int = 512
    n_rnn: int = 2
    q_levels: int = 256
    embedding_dim: int = 256
    mlp_dim: int = 512

    def __post_init__(self):
        nn.Module.__init__(self)
        GeneratingNetwork.__init__(self)
        n_tiers = len(self.frame_sizes)
        tiers = []
        if n_tiers > 2:
            for i, fs in enumerate(self.frame_sizes[:-2]):
                tiers += [SampleRNNTier(i, fs, self.dim,
                                        up_sampling=fs // self.frame_sizes[i + 1],
                                        n_rnn=self.n_rnn,
                                        q_levels=self.q_levels,
                                        embedding_dim=None,
                                        mlp_dim=None)]
        # before last tier
        tiers += [SampleRNNTier(n_tiers - 2, self.frame_sizes[-2], self.dim,
                                up_sampling=self.frame_sizes[-1],
                                n_rnn=self.n_rnn,
                                q_levels=self.q_levels,
                                embedding_dim=None,
                                mlp_dim=None)]
        # bottom tier
        tiers += [SampleRNNTier(n_tiers - 1, self.frame_sizes[-1], self.dim,
                                up_sampling=1,
                                n_rnn=self.n_rnn,
                                q_levels=self.q_levels,
                                embedding_dim=self.embedding_dim,
                                mlp_dim=self.mlp_dim)]

        self.tiers = nn.ModuleList(tiers)
        self.hidden = [None] * n_tiers

    def forward(self, tiers_inputs):
        prev_out = None
        for i, (tier, inpt) in enumerate(zip(self.tiers, tiers_inputs)):
            prev_out, self.hidden[i] = tier(inpt, prev_out, self.hidden[i])
        return prev_out

    def reset_h0(self):
        self.hidden = [None] * len(self.frame_sizes)

    def generate_(self, prompt, n_steps, temperature=1.):
        output = self.prepare_prompt(prompt, n_steps, at_least_nd=2)
        # trim to start with a whole number of top frames
        output = output[:, prompt.size(1) % self.frame_sizes[0]:]
        prior_t = prompt.size(1) - (prompt.size(1) % self.frame_sizes[0])

        # init variables
        fs = [*self.frame_sizes]
        outputs = [None] * (len(fs) - 1)
        # hidden are reset if prompt.size(0) != self.hidden.size(0)
        hiddens = self.hidden
        tiers = self.tiers

        for t in self.generate_tqdm(range(fs[0], n_steps + prior_t)):
            for i in range(len(tiers) - 1):
                if t % fs[i] == 0:
                    inpt = output[:, t - fs[i]:t].unsqueeze(1)

                    if i == 0:
                        prev_out = None
                    else:
                        prev_out = outputs[i - 1][:, (t // fs[i]) % (fs[i - 1] // fs[i])].unsqueeze(1)

                    out, h = tiers[i](inpt, prev_out, hiddens[i])
                    hiddens[i] = h
                    outputs[i] = out
            if t < prior_t:  # only used for warming-up
                continue
            prev_out = outputs[-1]
            inpt = output[:, t - fs[-1]:t].reshape(-1, 1, fs[-1])

            out, _ = tiers[-1](inpt, prev_out[:, (t % fs[-1]) - fs[-1]].unsqueeze(1))
            if temperature is None:
                pred = (nn.Softmax(dim=-1)(out.squeeze(1))).argmax(dim=-1)
            else:
                # great place to implement dynamic cooling/heating !
                pred = torch.multinomial(nn.Softmax(dim=-1)(out.squeeze(1) / temperature), 1)
            output.data[:, t] = pred.squeeze()

        return output

